import click
from typing import Optional
import logging
import pandas as pd
import numpy as np
import math
import pysam
from sklearn.preprocessing import MinMaxScaler
from mpralib.mpradata import MPRABarcodeData, BarcodeFilter, Modality
from mpralib.utils.io import (
    chromosome_map,
    export_activity_file,
    export_barcode_file,
    export_counts_file,
    read_sequence_design_file,
)
import mpralib.utils.plot as plt
from mpralib.utils.file_validation import validate_tsv_with_schema, ValidationSchema

pd.options.mode.copy_on_write = True


@click.group(help="Command line interface of MPRAlib, a library for MPRA data analysis.")
def cli() -> None:
    pass


@cli.group(help="Validate standardized MPRA reporter formats.")
def validate_file() -> None:
    pass


@validate_file.command(help="Validate MPRA reporter sequence design file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Sequence Design file to validate.",
)
def reporter_sequence_design(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_SEQUENCE_DESIGN):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate MPRA Reporter Barcode to Element Mapping file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Barcode to Element Mapping file to validate.",
)
def reporter_barcode_to_element_mapping(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_BARCODE_TO_ELEMENT_MAPPING):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Experiment Barcode file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Experiment Barcode file to validate.",
)
def reporter_experiment_barcode(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_EXPERIMENT_BARCODE):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Experiment file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Experiment file to validate.",
)
def reporter_experiment(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_EXPERIMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Element file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Element file to validate.",
)
def reporter_element(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_ELEMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Variant file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Variant file to validate.",
)
def reporter_variant(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_VARIANT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Genomic Element file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Genomic Element file to validate.",
)
def reporter_genomic_element(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_GENOMIC_ELEMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Genomic Variant file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Genomic Variant file to validate.",
)
def reporter_genomic_variant(input_file: str) -> None:
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_GENOMIC_VARIANT):
        raise click.ClickException("Validation failed. Please check the input file.")


@cli.group(help="General functionality.")
def functional() -> None:
    pass


@functional.command(help="Generating element activity or barcode count files.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA result in a barcode format.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output (element level only).",
)
@click.option(
    "--element-level/--barcode-level",
    "element_level",
    default=True,
    help="Export activity at the element (default) or barcode level.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of results.",
)
def activities(input_file: str, bc_threshold: int, element_level: bool, output_file: str) -> None:
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if element_level:
        export_activity_file(mpradata.oligo_data, output_file)
    else:
        export_barcode_file(mpradata, output_file)


@functional.command(help="Compute pairwise correlations under a certain barcode threshold.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA result in a barcode format.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output (element level only).",
)
@click.option(
    "--correlation-method",
    "correlation_method",
    required=False,
    default="all",
    type=click.Choice(["pearson", "spearman", "all"]),
    help="Computing pearson, spearman or both.",
)
@click.option(
    "--correlation-on",
    "correlation_on",
    required=False,
    default="all",
    type=click.Choice(["dna_normalized", "rna_normalized", "activity", "all"]),
    help="Using a barcode threshold for output (element level only).",
)
def compute_correlation(input_file: str, bc_threshold: int, correlation_on: str, correlation_method: str) -> None:
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    mpradata.barcode_threshold = bc_threshold

    if correlation_on == "all":
        correlation_on_mode = [Modality.DNA_NORMALIZED, Modality.RNA_NORMALIZED, Modality.ACTIVITY]
    else:
        correlation_on_mode = [Modality.from_string(correlation_on)]

    if correlation_method == "all":
        correlation_method_list = ["pearson", "spearman"]
    else:
        correlation_method_list = [correlation_method]

    for method in correlation_method_list:
        for mode in correlation_on_mode:
            click.echo(f"{method} correlation on {mode}: {mpradata.correlation(method, mode).flatten()[[1, 2, 5]]}")


functional.command(help="Filter out outliers based on RNA z-score and copute correlations before and afterwards.")


@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--outlier-rna-zscore-times",
    "rna_zscore_times",
    default=3,
    type=float,
    help="Absolute rna z_score is not allowed to be larger than this value.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output",
    "output_file",
    required=False,
    type=click.Path(writable=True),
    help="Output file of results.",
)
def filter_outliers(input_file, rna_zscore_times, bc_threshold, output_file):
    """Filters outliers from MPRA barcode data based on RNA z-score and barcode count threshold.

    Reads an input file to create an MPRAdata object, applies a barcode filter to remove outliers
    using the specified RNA z-score threshold, and optionally exports the filtered activity data
    to an output file. Prints Pearson correlation of log2FoldChange before and after outlier removal.

    Args:
        input_file (str): Path to the input file containing barcode data.
        rna_zscore_times (float): Number of standard deviations for RNA z-score outlier filtering.
        bc_threshold (int): Minimum barcode count threshold for filtering.
        output_file (str): Path to the output file to export filtered activity data. If None, no file is written.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    oligo_data = mpradata.oligo_data

    click.echo(
        f"""Pearson correlation log2FoldChange BEFORE outlier removal: {
            oligo_data.correlation('pearson', Modality.ACTIVITY).flatten()[[1, 2, 5]]
        }"""
    )

    mpradata.apply_barcode_filter(BarcodeFilter.RNA_ZSCORE, {"times_zscore": rna_zscore_times})

    oligo_data = mpradata.oligo_data
    click.echo(
        f"""Pearson correlation log2FoldChange AFTER outlier removal: {
            oligo_data.correlation('pearson', Modality.ACTIVITY).flatten()[[1, 2, 5]]
        }"""
    )
    if output_file:
        export_activity_file(oligo_data, output_file)


@cli.group(help="MPRA sequence design file functionality.")
def sequence_design():
    pass


@sequence_design.command(help="Using a metadata file to generate a oligo to variant mapping.")
@click.option(
    "--input",
    "input_file",
    required=False,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results (barcode output file). "
    "If set only map of present oligos in the barcode count file will be generated.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of results.",
)
def get_variant_map(input_file: str, sequence_design_file: str, output_file: str) -> None:

    df_sequence_design = read_sequence_design_file(sequence_design_file)

    if input_file:
        print("Read barcode count file...")
        mpradata = MPRABarcodeData.from_file(input_file)

        print("Generating oligo data...")
        mpradata = mpradata.oligo_data

        print("Adding metadata file...")
        mpradata.add_sequence_design(df_sequence_design, sequence_design_file)

        print(mpradata.data.var["SPDI"])
        print(type(mpradata.data.var["SPDI"]))
        print(mpradata.data.var["SPDI"].values)

        print("Generating variant map...")
        variant_map = mpradata.variant_map
    else:
        variant_map = pd.DataFrame(
            {
                "ID": np.concatenate(df_sequence_design["SPDI"].values),
                "allele": np.concatenate(df_sequence_design["allele"].values),
                "oligo": df_sequence_design.index.values.repeat(df_sequence_design["SPDI"].apply(lambda x: len(x))),
            }
        )

        variant_map["REF"] = np.where(
            variant_map["allele"] == "ref", variant_map["oligo"], np.full(len(variant_map["allele"]), None, dtype=object)
        )
        variant_map["ALT"] = np.where(
            variant_map["allele"] == "alt", variant_map["oligo"], np.full(len(variant_map["allele"]), None, dtype=object)
        )
        variant_map = variant_map.groupby("ID").agg(
            {"REF": lambda x: list(filter(None, x)), "ALT": lambda x: list(filter(None, x))}
        )

    print("Prepair output file...")
    for key in ["REF", "ALT"]:
        # TODO: what happens a variant has multiple alt alleles?
        # Right now joined by comma. But maybe I hould use one row per ref/alt pai?
        variant_map.loc[:, key] = variant_map[key].apply(lambda x: ",".join(x))

    variant_map = variant_map[variant_map.apply(lambda x: all(x != ""), axis=1)]

    variant_map.to_csv(output_file, sep="\t", index=True)


@sequence_design.command(help="Return DNA and RNA counts for all oligos or only thosed tagges as elements/ref snvs.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--normalized-counts/--counts",
    "normalized_counts",
    required=False,
    type=click.BOOL,
    default=False,
    help="Getting counts or normalized counts.",
)
@click.option(
    "--elements-only/--all-oligos",
    "elements_only",
    required=False,
    type=click.BOOL,
    default=False,
    help="Only return count data for elements and ref sequence of variants.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of all non zero counts.",
)
def get_counts(
    input_file: str,
    sequence_design_file: str,
    bc_threshold: int,
    normalized_counts: bool,
    use_oligos: bool,
    elements_only: bool,
    output_file: str,
) -> None:
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    element_mask = None
    if elements_only:

        element_mask = np.array(mpradata.data.var["allele"].apply(lambda x: "ref" in x).values, dtype=bool) | (
            mpradata.data.var["category"] == "element"
        )
        element_mask = ~np.repeat(np.array(element_mask)[np.newaxis, :], mpradata.n_obs, axis=0)

    export_counts_file(mpradata, output_file, normalized=normalized_counts, filter=element_mask)


@sequence_design.command(help="Write out DNA and RNA counts for REF and ALT oligos.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--normalized-counts/--counts",
    "normalized_counts",
    required=False,
    type=click.BOOL,
    default=False,
    help="Getting counts or normalized counts.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of all non zero counts, divided by ref and alt.",
)
def get_variant_counts(
    input_file: str, sequence_design_file: str, bc_threshold: int, normalized_counts: bool, use_oligos: bool, output_file: str
) -> None:
    """Processes MPRA (Massively Parallel Reporter Assay) data to generate variant-level count tables.

    Args:
        input_file (str): Path to the input file containing barcode count data.
        sequence_design_file (str): Path to the sequence design file mapping barcodes to variants/oligos.
        bc_threshold (int): Minimum barcode count threshold for filtering.
        normalized_counts (bool): If True, use normalized counts; otherwise, use raw counts.
        use_oligos (bool): If True, aggregate counts at the oligo level and map to variants; otherwise, export counts directly.
        output_file (str): Path to the output file where the variant count table will be saved.

    Returns:
        None. The function writes the resulting count table to the specified output file.

    Notes:
        - When `use_oligos` is True, the function aggregates DNA and RNA counts for reference and alternate oligos
          per variant, across all replicates, and outputs a table with these counts.
        - When `use_oligos` is False, the function exports the counts using `export_counts_file`.
        - Variants with all-zero counts or counts only on reference or alternate are removed from the output.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

        variant_map = mpradata.variant_map

        df = {"variant_id": []}
        for replicate in mpradata.obs_names:
            for rna_or_dna in ["dna", "rna"]:
                df[f"{rna_or_dna}_count_{replicate}_REF"] = []
                df[f"{rna_or_dna}_count_{replicate}_ALT"] = []

        if normalized_counts:
            dna_counts = mpradata.normalized_dna_counts.copy()
            rna_counts = mpradata.normalized_rna_counts.copy()
        else:
            dna_counts = mpradata.dna_counts.copy()
            rna_counts = mpradata.rna_counts.copy()

        not_observed_mask = ~mpradata.observed
        lower_bc_mask = mpradata.barcode_counts < mpradata.barcode_threshold
        not_variant_mask = ~np.repeat(
            np.array(mpradata.data.var["category"] == "variant")[np.newaxis, :], mpradata.n_obs, axis=0
        )

        dna_counts = np.ma.masked_array(dna_counts, mask=np.any([not_observed_mask, lower_bc_mask, not_variant_mask], axis=0))
        rna_counts = np.ma.masked_array(rna_counts, mask=np.any([not_observed_mask, lower_bc_mask, not_variant_mask], axis=0))

        for spdi, row in variant_map.iterrows():

            mask_ref = mpradata.oligos.isin(row["REF"])
            mask_alt = mpradata.oligos.isin(row["ALT"])

            df["variant_id"].append(spdi)

            dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
            rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

            dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
            rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

            for idx, replicate in enumerate(mpradata.obs_names):
                df[f"dna_count_{replicate}_REF"].append(dna_counts_ref[idx])
                df[f"rna_count_{replicate}_REF"].append(rna_counts_ref[idx])
                df[f"dna_count_{replicate}_ALT"].append(dna_counts_alt[idx])
                df[f"rna_count_{replicate}_ALT"].append(rna_counts_alt[idx])

        df = pd.DataFrame(df).set_index("variant_id")
        # remove IDs which are all zero or just on ref or alt
        df = df.map(lambda x: np.nan if isinstance(x, np.ma.core.MaskedConstant) else x)
        ref_columns = [f"dna_count_{replicate}_REF" for replicate in mpradata.obs_names] + [
            f"rna_count_{replicate}_REF" for replicate in mpradata.obs_names
        ]
        alt_columns = [f"dna_count_{replicate}_ALT" for replicate in mpradata.obs_names] + [
            f"rna_count_{replicate}_ALT" for replicate in mpradata.obs_names
        ]
        nan_columns = df.loc[:, ref_columns].isna().all(axis=1) | df.loc[:, alt_columns].isna().all(axis=1)
        df = df[~nan_columns]

        if normalized_counts:
            df.to_csv(output_file, sep="\t", index=True, na_rep="", float_format="%.6f")
        else:
            df.to_csv(output_file, sep="\t", index=True, na_rep="", float_format="%.0f")
    else:
        export_counts_file(
            mpradata,
            output_file,
            normalized=normalized_counts,
            filter=~np.repeat(np.array(mpradata.data.var["category"] == "variant")[np.newaxis, :], mpradata.n_obs, axis=0),
        )


@sequence_design.command()
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of mpralm or BCalm results.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output-reporter-elements",
    "output_reporter_elements_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_elements(
    input_file: str, sequence_design_file: str, statistics_file: str, bc_threshold: int, output_reporter_elements_file: str
) -> None:
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    df = pd.read_csv(statistics_file, sep="\t", header=0)

    indexes_in_order = [mpradata.oligos[mpradata.oligos == ID].index.tolist() for ID in df["ID"]]
    indexes_in_order = [index for sublist in indexes_in_order for index in sublist]
    df.index = pd.Index(indexes_in_order)

    df = df.join(mpradata.oligos, how="right")

    mpradata.data.varm["mpralm_element"] = df

    out_df = df[["oligo", "logFC", "P.Value", "adj.P.Val"]]
    out_df.loc[:, "inputCount"] = mpradata.normalized_dna_counts.mean(axis=0)
    out_df.loc[:, "outputCount"] = mpradata.normalized_rna_counts.mean(axis=0)
    out_df.dropna(inplace=True)
    out_df.loc[out_df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    out_df.loc[out_df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    out_df.loc[:, "minusLog10PValue"] = np.abs(np.log10(out_df.loc[:, "P.Value"]))
    out_df.loc[:, "minusLog10QValue"] = np.abs(np.log10(out_df.loc[:, "adj.P.Val"]))
    out_df = out_df.rename(columns={"oligo": "oligo_name", "logFC": "log2FoldChange"})
    out_df[
        [
            "oligo_name",
            "log2FoldChange",
            "inputCount",
            "outputCount",
            "minusLog10PValue",
            "minusLog10QValue",
        ]
    ].to_csv(output_reporter_elements_file, sep="\t", index=False, float_format="%.4f")


@sequence_design.command()
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of the result from mpralm or BCalm.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output-reporter-variants",
    "output_reporter_variants_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_variants(
    input_file: str, sequence_design_file: str, statistics_file: str, bc_threshold: int, output_reporter_variants_file: str
) -> None:
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    variant_map = mpradata.variant_map

    df = pd.read_csv(statistics_file, sep="\t", header=0, index_col=0, na_values="NA").dropna()

    dna_counts = mpradata.normalized_dna_counts.copy()
    rna_counts = mpradata.normalized_rna_counts.copy()

    for spdi, row in variant_map.iterrows():

        mask_ref = mpradata.oligos.isin(row["REF"])
        mask_alt = mpradata.oligos.isin(row["ALT"])

        dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
        rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

        dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
        rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

        if spdi in df.index:
            df.loc[str(spdi), "inputCountRef"] = dna_counts_ref.mean()
            df.loc[str(spdi), "inputCountAlt"] = dna_counts_alt.mean()
            df.loc[str(spdi), "outputCountRef"] = rna_counts_ref.mean()
            df.loc[str(spdi), "outputCountAlt"] = rna_counts_alt.mean()
            df.loc[str(spdi), "variantPos"] = int(
                mpradata.data.var["variant_pos"][mpradata.oligos.isin(row["REF"])].values[0][0]
            )

    df.loc[df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    df.loc[df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    df["minusLog10PValue"] = np.abs(np.log10(df["P.Value"]))
    df["minusLog10QValue"] = np.abs(np.log10(df["adj.P.Val"]))
    df.rename(
        columns={
            "CI.L": "CI_lower_95",
            "CI.R": "CI_upper_95",
            "logFC": "log2FoldChange",
        },
        inplace=True,
    )
    df["postProbEffect"] = df["B"].apply(lambda x: math.exp(x) / (1 + math.exp(x)))
    df["variant_id"] = df.index
    df["refAllele"] = df["variant_id"].apply(lambda x: x.split(":")[2])
    df["altAllele"] = df["variant_id"].apply(lambda x: x.split(":")[3])
    df["variantPos"] = df["variantPos"].astype(int)

    df[
        [
            "variant_id",
            "log2FoldChange",
            "inputCountRef",
            "outputCountRef",
            "inputCountAlt",
            "outputCountAlt",
            "minusLog10PValue",
            "minusLog10QValue",
            "postProbEffect",
            "CI_lower_95",
            "CI_upper_95",
            "variantPos",
            "refAllele",
            "altAllele",
        ]
    ].to_csv(output_reporter_variants_file, sep="\t", index=False, float_format="%.4f")


@sequence_design.command()
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA lm file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--reference",
    "reference",
    required=True,
    type=str,
    help="Using only this reference as denoted in ref in the sequence design file.",
)
@click.option(
    "--output-reporter-genomic-elements",
    "output_reporter_genomic_elements_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_genomic_elements(
    input_file: str,
    sequence_design_file: str,
    statistics_file: str,
    bc_threshold: int,
    reference: str,
    output_reporter_genomic_elements_file: str,
) -> None:

    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    mask = np.array(mpradata.data.var["allele"].apply(lambda x: "ref" in x).values, dtype=bool) | (
        mpradata.data.var["category"] == "element"
    )
    mask = mask & (mpradata.data.var["ref"] == reference)

    df = pd.read_csv(statistics_file, sep="\t", header=0)

    indexes_in_order = [mpradata.oligos[mpradata.oligos == ID].index.tolist() for ID in df["ID"]]
    indexes_in_order = [index for sublist in indexes_in_order for index in sublist]
    df.index = pd.Index(indexes_in_order)

    df = df.join(mpradata.oligos, how="right")

    mpradata.data.varm["mpralm_element"] = df

    out_df = df[["oligo", "logFC", "P.Value", "adj.P.Val"]]

    out_df.loc[:, ["inputCount"]] = mpradata.normalized_dna_counts.mean(axis=0)
    out_df.loc[:, ["outputCount"]] = mpradata.normalized_rna_counts.mean(axis=0)

    out_df["chr"] = mpradata.data.var["chr"]
    out_df["start"] = mpradata.data.var["start"]
    out_df["end"] = mpradata.data.var["end"]
    out_df["strand"] = mpradata.data.var["strand"]

    # apply ref and element mask
    out_df = out_df[mask]

    out_df.dropna(inplace=True)
    scaler = MinMaxScaler(feature_range=(0, 1000))
    out_df["score"] = scaler.fit_transform(np.abs(out_df[["logFC"]])).astype(int)
    out_df.loc[out_df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    out_df.loc[out_df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    out_df["minusLog10PValue"] = np.abs(np.log10(out_df["P.Value"]))
    out_df["minusLog10QValue"] = np.abs(np.log10(out_df["adj.P.Val"]))
    out_df.rename(columns={"oligo": "name", "logFC": "log2FoldChange"}, inplace=True)
    out_df = out_df[
        [
            "chr",
            "start",
            "end",
            "name",
            "score",
            "strand",
            "log2FoldChange",
            "inputCount",
            "outputCount",
            "minusLog10PValue",
            "minusLog10QValue",
        ]
    ].sort_values(by=["chr", "start", "end"])

    with pysam.BGZFile(output_reporter_genomic_elements_file, "ab", None) as f:
        f.write(out_df.to_csv(sep="\t", index=False, header=False, float_format="%.4f").encode())


@sequence_design.command()
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of the result from mpralm or BCalm.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--reference",
    "reference",
    required=True,
    default="GRCh38",
    type=click.Choice(["GRCh38", "GRCh37"]),
    help="Using only this reference as denoted in ref in the sequence design file.",
)
@click.option(
    "--output-reporter-genomic-variants",
    "output_reporter_genomic_variants_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_genomic_variants(
    input_file: str,
    sequence_design_file: str,
    statistics_file: str,
    bc_threshold: int,
    reference: str,
    output_reporter_genomic_variants_file: str,
) -> None:

    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    variant_map = mpradata.variant_map

    df = pd.read_csv(statistics_file, sep="\t", header=0, index_col=0, na_values="NA").dropna()

    dna_counts = mpradata.normalized_dna_counts.copy()
    rna_counts = mpradata.normalized_rna_counts.copy()

    for spdi, row in variant_map.iterrows():

        mask_ref = mpradata.oligos.isin(row["REF"])
        mask_alt = mpradata.oligos.isin(row["ALT"])

        dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
        rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

        dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
        rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

        if spdi in df.index:
            df.loc[str(spdi), "inputCountRef"] = dna_counts_ref.mean()
            df.loc[str(spdi), "inputCountAlt"] = dna_counts_alt.mean()
            df.loc[str(spdi), "outputCountRef"] = rna_counts_ref.mean()
            df.loc[str(spdi), "outputCountAlt"] = rna_counts_alt.mean()
            df.loc[str(spdi), "variantPos"] = int(
                mpradata.data.var["variant_pos"][mpradata.oligos.isin(row["REF"])].values[0][0]
            )
            df.loc[str(spdi), "strand"] = mpradata.data.var["strand"][mpradata.oligos.isin(row["REF"])].values[0]

    df["variantPos"] = df["variantPos"].astype(int)
    df.loc[df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    df.loc[df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    df["minusLog10PValue"] = np.abs(np.log10(df["P.Value"]))
    df["minusLog10QValue"] = np.abs(np.log10(df["adj.P.Val"]))
    df.rename(
        columns={
            "CI.L": "CI_lower_95",
            "CI.R": "CI_upper_95",
            "logFC": "log2FoldChange",
        },
        inplace=True,
    )
    df["postProbEffect"] = df["B"].apply(lambda x: math.exp(x) / (1 + math.exp(x)))
    df["variant_id"] = df.index
    df["refAllele"] = df["variant_id"].apply(lambda x: x.split(":")[2])
    df["altAllele"] = df["variant_id"].apply(lambda x: x.split(":")[3])
    df["start"] = df["variant_id"].apply(lambda x: x.split(":")[1]).astype(int)
    df["end"] = df["start"] + df["refAllele"].apply(lambda x: len(x)).astype(int)

    map = chromosome_map()
    map = map[map["release"] == reference]
    df["chr"] = df["variant_id"].apply(lambda x: _get_chr(map, x, mpradata.LOGGER))

    df.dropna(inplace=True)

    scaler = MinMaxScaler(feature_range=(0, 1000))
    df["score"] = scaler.fit_transform(np.abs(df[["log2FoldChange"]])).astype(int)

    df = df[
        [
            "chr",
            "start",
            "end",
            "variant_id",
            "score",
            "strand",
            "log2FoldChange",
            "inputCountRef",
            "outputCountRef",
            "inputCountAlt",
            "outputCountAlt",
            "minusLog10PValue",
            "minusLog10QValue",
            "postProbEffect",
            "CI_lower_95",
            "CI_upper_95",
            "variantPos",
            "refAllele",
            "altAllele",
        ]
    ].sort_values(by=["chr", "start", "end"])

    with pysam.BGZFile(output_reporter_genomic_variants_file, "ab", None) as f:
        f.write(df.to_csv(sep="\t", index=False, header=False, float_format="%.4f").encode())


def _get_chr(map: pd.DataFrame, variant_id: str, logger: logging.Logger) -> Optional[str]:
    variant_contig = variant_id.split(":")[0]
    if variant_contig in map["refseq"].values:
        return map[map["refseq"] == variant_contig].loc[:, "ucsc"].values[0]
    else:
        logger.warning(f"Contig {variant_contig} of SPDI {variant_id} not found in chromosome map. Returning None.")
        return None


@cli.group(help="Plotting functions.")
def plot() -> None:
    pass


@plot.command()
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--modality",
    "modality",
    required=False,
    default="activity",
    type=click.Choice(["dna_normalized", "rna_normalized", "activity"]),
    help="What modality should be plotted.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Copare only these two replicates. Ottheriwse all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def correlation(
    input_file: str, use_oligos: bool, bc_threshold: int, modality: str, replicates: Optional[list], output_file
) -> None:
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    mod = Modality.from_string(modality)

    if replicates:
        fig = plt.correlation(mpradata, mod, replicates)
    else:
        fig = plt.correlation(mpradata, mod)

    fig.savefig(output_file)


@plot.command(help="Plotting the DNA vs RNA counts (log10, median on  multiple replicates).")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Copare only these two replicates. Ottheriwse all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def dna_vs_rna(input_file, use_oligos, bc_threshold, replicates, output_file):
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    if replicates:
        fig = plt.dna_vs_rna(mpradata, replicates)
    else:
        fig = plt.dna_vs_rna(mpradata)

    fig.savefig(output_file)


@plot.command(help="Histogramm of barcodes per oligo.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Show only these replicates. Otheriwse all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def barcodes_per_oligo(input_file: str, replicates: Optional[list], output_file: str) -> None:
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    if replicates:
        fig = plt.barcodes_per_oligo(mpradata, replicates)
    else:
        fig = plt.barcodes_per_oligo(mpradata)

    fig.savefig(output_file)


@plot.command(help="Outlier plot.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
def outlier(input_file: str, bc_threshold: int) -> None:
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    plt.barcodes_outlier(mpradata)

    # fig.savefig(output_file)


if __name__ == "__main__":
    cli()



from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
import anndata as ad
from scipy.stats import spearmanr, pearsonr
from enum import Enum
import logging
import os
from mpralib.exception import MPRAlibException
from typing import Any, Optional
from numpy.typing import NDArray


class Modality(Enum):
    """An enumeration representing different data modalities in MPRA (Massively Parallel Reporter Assay) experiments."""

    DNA = "DNA"
    """str: Represents DNA data modality. """
    RNA = "RNA"
    """str: Represents RNA data modality."""
    DNA_NORMALIZED = "DNA_NORMALIZED"
    """str: Represents normalized DNA data modality."""
    RNA_NORMALIZED = "RNA_NORMALIZED"
    """str: Represents normalized RNA data modality."""
    ACTIVITY = "ACTIVITY"
    """str: Represents activity data modality, typically calculated as the log2 ratio of normalized RNA to DNA counts."""

    def __new__(cls, value):
        obj = object.__new__(cls)
        obj._value_ = str(value).lower()
        return obj

    @classmethod
    def from_string(cls, value: str) -> "Modality":
        """Creates a Modality enum member from a string value.

        Args:
            value (str): The string representation of the enum member.

        Returns:
            The corresponding Modality enum member.

        Raises:
            ValueError: If the provided string does not match any Modality member.
        """
        for member in cls:
            if member.value == value.lower():
                return member
        raise ValueError(f"{value} is not a valid {cls.__name__}")


class CountSampling(Enum):
    """Enumeration representing the types of count sampling available for MPRA data."""

    RNA = "RNA"
    """str: Represents RNA count sampling."""
    DNA = "DNA"
    """str: Represents DNA count sampling."""
    RNA_AND_DNA = "RNA_AND_DNA"
    """str: Represents both RNA and DNA count sampling."""


class BarcodeFilter(Enum):
    """Enumeration of available barcode filtering methods."""

    RNA_ZSCORE = "RNA_ZSCORE"
    """str: Filter barcodes based on RNA z-score."""
    MAD = "MAD"
    """str: Filter barcodes based on Median Absolute Deviation (MAD)."""
    RANDOM = "RANDOM"
    """str: Randomly filter barcodes."""
    MIN_COUNT = "MIN_COUNT"
    """str: Filter barcodes with counts below a specified minimum."""
    MAX_COUNT = "MAX_COUNT"
    """str: Filter barcodes with counts above a specified maximum."""


class MPRAData(ABC):
    """Abstract base class for handling MPRA (Massively Parallel Reporter Assay) data using AnnData objects.

    This class provides a standardized interface and core functionality for managing, normalizing filtering, and analyzing MPRA data, including DNA/RNA counts, barcode handling, activity computation, and correlation analysis. It is designed to be subclassed for specific MPRA data formats.

    Args:
        data (anndata.AnnData): The AnnData object containing MPRA data.
        barcode_threshold (int, optional): Minimum barcode count threshold for filtering. Defaults to 0.

    Attributes:
        _SCALING (float): Default scaling factor for normalization.
        _PSEUDOCOUNT (int): Default pseudocount for normalization.
        _data (anndata.AnnData): The AnnData object containing MPRA data.

    Raises:
        ValueError: If required metadata (e.g., sequence design file) is not loaded.
    """  # noqa: E501

    LOGGER = logging.getLogger(__name__)
    """logging.Logger: Logger for the class."""
    LOGGER.setLevel(logging.INFO)

    _SCALING = 1e6
    _PSEUDOCOUNT = 1

    @classmethod
    @abstractmethod
    def from_file(cls, file_path: str) -> "MPRAData":
        """Create an instance of the class from a file.

        This method reads data from a specified file (reporter experiment barcode or reporter experiment file format), processes it, and returns an instance of the class containing the data in an AnnData object.

        Args:
            file_path (str): Path to the input file containing reporter experiment barcode or reporter experiment data.

        Returns:
            An instance of MPRAData containing the processed data in an AnnData object.

        Raises:
            IOError: If the file cannot be read or parsed.
            ValueError: If the file format is invalid.
        """  # noqa: E501
        pass

    def __init__(self, data: ad.AnnData, barcode_threshold: int = 0):
        self._data = data
        self.var_filter = None
        self.barcode_threshold = barcode_threshold

        # Initialize scaling and pseudo count metadata
        scaling = self._get_metadata("SCALING")
        if scaling is not None:
            self._SCALING = scaling
        else:
            self._add_metadata("SCALING", self._SCALING)

        pseudo_count = self._get_metadata("PSEUDOCOUNT")
        if pseudo_count is not None:
            self._PSEUDOCOUNT = pseudo_count
        else:
            self._add_metadata("PSEUDOCOUNT", self._PSEUDOCOUNT)

    @property
    def scaling(self) -> float:
        """float: Scaling factor for normalization."""
        return self._SCALING

    @scaling.setter
    def scaling(self, new_scaling: float) -> None:
        if new_scaling != self._SCALING:
            self.drop_normalized()
            self._SCALING = new_scaling
            self._add_metadata("SCALING", self._SCALING)

    @property
    def pseudo_count(self) -> int:
        """int: Pseudocount added during normalization to avoid division by zero."""
        return self._PSEUDOCOUNT

    @pseudo_count.setter
    def pseudo_count(self, new_pseudo_count: int) -> None:
        if new_pseudo_count != self._PSEUDOCOUNT:
            self.drop_normalized()
            self._PSEUDOCOUNT = new_pseudo_count
            self._add_metadata("PSEUDOCOUNT", self._PSEUDOCOUNT)

    @property
    def data(self) -> ad.AnnData:
        """ad.AnnData: The underlying AnnData object containing MPRA data."""
        return self._data

    @data.setter
    def data(self, new_data: ad.AnnData) -> None:
        self._data = new_data

    @property
    def var_names(self) -> pd.Index:
        """pd.Index: Returns the variable names (samples) of the dataset."""
        return self.data.var_names

    @property
    def n_vars(self) -> int:
        """int: Returns the number of variables (samples) in the dataset."""
        return self.data.n_vars

    @property
    def obs_names(self) -> pd.Index:
        """pd.Index: Returns the observation names (barcodes) of the dataset."""
        return self.data.obs_names

    @property
    def n_obs(self) -> int:
        """int: Returns the number of observations (barcodes) in the dataset."""
        return self.data.n_obs

    @property
    def oligos(self) -> pd.Series:
        """pd.Series: Returns the oligo names for each variable in the dataset."""
        return self.data.var["oligo"]

    @property
    def raw_rna_counts(self) -> NDArray[np.int32]:
        """NDArray[np.int32]: Returns the raw RNA counts from the dataset."""
        return np.asarray(self.data.layers["rna"])

    @property
    def normalized_dna_counts(self) -> NDArray[np.float32]:
        """NDArray[np.float32]: Returns the normalized DNA counts from the dataset."""
        if "dna_normalized" not in self.data.layers:
            self._normalize()
        return np.asarray(self.data.layers["dna_normalized"])

    @property
    def rna_counts(self) -> NDArray[np.int32]:
        """NDArray[np.int32]: Returns the RNA counts, applying the variable filter if present."""
        if "rna_sampling" in self.data.layers:
            return np.asarray(self.data.layers["rna_sampling"]) * ~self.var_filter.T
        else:
            return self.raw_rna_counts * ~self.var_filter.T

    @property
    def raw_dna_counts(self) -> NDArray[np.int32]:
        """NDArray[np.int32]: Returns the raw DNA counts from the dataset."""
        return np.asarray(self.data.layers["dna"])

    @property
    def dna_counts(self) -> NDArray[np.int32]:
        """NDArray[np.int32]: Returns the DNA counts, applying the variable filter if present."""
        if "dna_sampling" in self.data.layers:
            return np.asarray(self.data.layers["dna_sampling"]) * ~self.var_filter.T
        else:
            return self.raw_dna_counts * ~self.var_filter.T

    @property
    def normalized_rna_counts(self) -> NDArray[np.float32]:
        """NDArray[np.float32]: Returns the normalized RNA counts from the dataset."""
        if "rna_normalized" not in self.data.layers:
            self._normalize()
        return np.asarray(self.data.layers["rna_normalized"])

    @property
    def activity(self) -> NDArray[np.float32]:
        """NDArray[np.float32]: Returns the activity values calculated from normalized RNA and DNA counts."""
        if "activity" not in self.data.layers:
            self._compute_activities()
        return np.asarray(self.data.layers["activity"])

    def _compute_activities(self) -> None:
        ratio = np.divide(
            self.normalized_rna_counts,
            self.normalized_dna_counts,
            where=self.normalized_dna_counts != 0,
        )
        with np.errstate(divide="ignore"):
            log2ratio = np.log2(ratio)
            log2ratio[np.isneginf(log2ratio)] = np.nan
        self.data.layers["activity"] = log2ratio * ~self.var_filter.T

    @property
    def observed(self) -> NDArray[np.bool_]:
        """:class:`NDArray[np.bool_]`: Returns a boolean NumPy array indicating which barcodes (observations) have non-zero counts in either DNA or RNA."""  # noqa: E501

        return (self.dna_counts + self.rna_counts) > 0

    @property
    def var_filter(self) -> NDArray[np.bool_]:
        """:class:`NDArray[np.bool_]`: Returns a boolean NumPy array indicating which variables (samples) are filtered out."""

        return np.asarray(self.data.varm["var_filter"], dtype=np.bool_)

    @var_filter.setter
    def var_filter(self, new_data: Optional[NDArray[np.bool_]]) -> None:
        if new_data is None:
            self.data.varm["var_filter"] = np.full((self.data.n_vars, self.data.n_obs), False)
            if "var_filter" in self.data.uns:
                del self.data.uns["var_filter"]
        else:
            self.data.varm["var_filter"] = new_data

        self.drop_barcode_counts()
        self.drop_normalized()

    @property
    def barcode_counts(self) -> NDArray[np.int32]:
        """NDArray[np.int32]: Returns the barcode counts matrix, which is a sum of counts across all barcodes for each oligo."""  # noqa: E501
        if "barcode_counts" not in self.data.layers or self.data.layers["barcode_counts"] is None:
            self.data.layers["barcode_counts"] = self._barcode_counts()
        return np.asarray(self.data.layers["barcode_counts"], dtype=np.int32)

    @barcode_counts.setter
    def barcode_counts(self, new_data: pd.DataFrame) -> None:
        self.data.layers["barcode_counts"] = new_data

    @abstractmethod
    def _barcode_counts(self) -> NDArray[np.int32]:
        pass

    @abstractmethod
    def drop_barcode_counts(self) -> None:
        """Removes or clears the barcode counts data from the current object.

        Raises:
            NotImplementedError: If the method is not yet implemented.
        """
        pass

    @property
    def barcode_threshold(self) -> int:
        """int: Returns the threshold for barcode filtering."""
        threshold = self._get_metadata("barcode_threshold")
        if threshold is None:
            # If no threshold is set, default to 0
            return 0
        else:
            return threshold

    @barcode_threshold.setter
    def barcode_threshold(self, barcode_threshold: int) -> None:
        self._add_metadata("barcode_threshold", barcode_threshold)
        self._drop_correlation()

    @property
    def variant_map(self) -> pd.DataFrame:
        """pd.DataFrame: Returns a DataFrame mapping SPDI IDs to alleles and oligos.

        Raises:
            ValueError: If the sequence design file is not loaded in the metadata.
        """
        if not self._get_metadata("sequence_design_file") and not (
            isinstance(self, MPRAOligoData) and self._get_metadata("MPRABarcodeData_sequence_design_file")
        ):
            raise ValueError("Sequence design file not loaded.")

        oligos = self.data.var["oligo"].repeat(self.data.var["SPDI"].apply(lambda x: len(x)).tolist())

        spdis = np.concatenate(self.data.var["SPDI"].values)

        alleles = np.concatenate(self.data.var["allele"].values)

        df = pd.DataFrame({"ID": spdis, "allele": alleles, "oligo": oligos})
        df["REF"] = df["oligo"].where(df["allele"] == "ref", None)
        df["ALT"] = df["oligo"].where(df["allele"] == "alt", None)
        df = df.groupby("ID").agg({"REF": lambda x: list(filter(None, x)), "ALT": lambda x: list(filter(None, x))})
        df = df[(df["REF"].apply(len) >= 1) & (df["ALT"].apply(len) >= 1)]

        return df

    @abstractmethod
    def _normalize(self) -> None:
        pass

    def drop_normalized(self) -> None:
        """Removes normalized RNA and DNA data layers from the dataset.

        This method deletes the "rna_normalized" and "dna_normalized" layers from the `self.data.layers` attribute, logs the operation, updates the metadata to indicate that normalization is no longer present, and drops any associated correlation data.
        """  # noqa: E501

        self.LOGGER.info("Dropping normalized data")

        self.data.layers.pop("rna_normalized", None)
        self.data.layers.pop("dna_normalized", None)
        self._drop_correlation()
        self._add_metadata("normalized", False)

    def correlation(self, method: str = "pearson", count_type: "Modality" = Modality.ACTIVITY) -> NDArray[np.float32]:
        """Calculates and return the correlation for activity or normalized counts.

        Returns:
            The Pearson or Spearman correlation matrix.
        """

        if count_type == Modality.DNA_NORMALIZED:
            filtered = self.normalized_dna_counts.copy()
            layer_name = str(count_type.value)
        elif count_type == Modality.RNA_NORMALIZED:
            layer_name = str(count_type.value)
            filtered = self.normalized_rna_counts.copy()
        elif count_type == Modality.ACTIVITY:
            filtered = self.activity.copy()
            layer_name = str(count_type.value)
        else:
            raise ValueError(f"Unsupported count type: {count_type}")

        filtered[self.barcode_counts < self.barcode_threshold] = np.nan

        return self._correlation(method, filtered, layer_name)

    def _correlation(self, method: str, data: NDArray[np.float32], layer: str) -> NDArray[np.float32]:
        if not self._get_metadata(f"correlation_{layer}"):
            self._compute_correlation(data, layer)
        correlation = self.data.obsp[f"{method}_correlation_{layer}"]
        correlation = np.asarray(correlation, dtype=np.float32)
        return correlation

    def _compute_correlation(self, data: NDArray[np.float32], layer) -> None:

        correlation_methods = ["pearson", "spearman"]

        def compute_correlation(x, y, method: str) -> tuple:
            if method == "spearman":
                return spearmanr(x, y)
            elif method == "pearson":
                return pearsonr(x, y)
            else:
                raise ValueError(f"Unsupported correlation method: {method}")

        # apply var filter to data
        data[self.var_filter.T] = np.nan

        num_columns = self.n_obs

        correlations = {}
        pvalues = {}
        for method in correlation_methods:
            correlations[method] = np.zeros((num_columns, num_columns), dtype=np.float32)
            pvalues[method] = np.zeros((num_columns, num_columns), dtype=np.float32)

        for i in range(num_columns):
            for j in range(i, num_columns):
                mask = ~np.isnan(data[i, :]) & ~np.isnan(data[j, :])
                x = data[i, mask]
                y = data[j, mask]

                for method in correlation_methods:
                    corr, pvalue = compute_correlation(x, y, method)
                    correlations[method][i, j] = corr
                    pvalues[method][i, j] = pvalue
                    correlations[method][j, i] = corr
                    pvalues[method][j, i] = pvalue

        for method in correlation_methods:
            self.data.obsp[f"{method}_correlation_{layer}"] = correlations[method]
            self.data.obsp[f"{method}_correlation_{layer}_pvalue"] = pvalues[method]

        self._add_metadata(f"correlation_{layer}", True)

    def _drop_correlation(self) -> None:

        for layer in ["rna_normalized", "dna_normalized", "activity"]:
            if self._get_metadata(f"correlation_{layer}"):
                self.LOGGER.info(f"Dropping correlation for layer {layer}")

                self._add_metadata(f"correlation_{layer}", False)
                for method in ["pearson", "spearman"]:
                    del self.data.obsp[f"{method}_correlation_{layer}"]
                    del self.data.obsp[f"{method}_correlation_{layer}_pvalue"]

    def _add_metadata(self, key: str, value: Any) -> None:
        if isinstance(value, list):
            if key not in self.data.uns:
                self.data.uns[key] = value
            else:
                self.data.uns[key] = self.data.uns[key] + value
        else:
            self.data.uns[key] = value

    def _get_metadata(self, key) -> Optional[Any]:
        if key in self.data.uns:
            return self.data.uns[key]
        else:
            return None

    def add_sequence_design(self, df_sequence_design: pd.DataFrame, sequence_design_file_path) -> None:
        """Add sequence design metadata to the object's data.

        Args:
            df_sequence_design (pd.DataFrame): DataFrame containing sequence design information, indexed by oligo identifiers.
            sequence_design_file_path (str): Path to the file from which the sequence design data was loaded to store it into the metadata.
        """  # noqa: E501

        df_matched_metadata = df_sequence_design.loc[self.oligos]

        self.data.var["category"] = pd.Categorical(df_matched_metadata["category"])
        self.data.var["class"] = pd.Categorical(df_matched_metadata["class"])
        self.data.var["ref"] = pd.Categorical(df_matched_metadata["ref"])
        self.data.var["chr"] = pd.Categorical(df_matched_metadata["chr"])
        self.data.var["start"] = df_matched_metadata["start"].values
        self.data.var["end"] = df_matched_metadata["end"].values
        self.data.var["strand"] = pd.Categorical(df_matched_metadata["strand"])
        self.data.var["variant_class"] = df_matched_metadata["variant_class"].values
        self.data.var["variant_pos"] = df_matched_metadata["variant_pos"].values
        self.data.var["SPDI"] = df_matched_metadata["SPDI"].values
        self.data.var["allele"] = df_matched_metadata["allele"].values

        self._add_metadata("sequence_design_file", sequence_design_file_path)

    def write(self, file_data_path: os.PathLike) -> None:
        """Writes the AnnData object to a file.

        Args:
            file_data_path (os.PathLike): The path where the AnnData object will be saved.
        """
        self.data.write(file_data_path)
        self.LOGGER.info(f"Data written to {file_data_path}")

    @classmethod
    def read(cls, file_data_path: str) -> "MPRAData":
        """Reads an AnnData object from a file.

        Args:
            file_data_path (str): The path from which the AnnData object will be read.

        Returns:
            MPRAData: An instance of the class containing the data read from the file.
        """
        data = ad.read_h5ad(file_data_path)
        return cls(data)


class MPRABarcodeData(MPRAData):
    """A class for handling barcode-level MPRA (Massively Parallel Reporter Assay) data, providing methods for data import, normalization, filtering, and aggregation to oligo-level data.

    This class extends `MPRAData` and is designed to work with barcode-resolved MPRA datasets, supporting a variety of barcode filtering strategies, normalization routines, and data transformations. It leverages AnnData for data storage and manipulation.

    Note:
        - Filtering and normalization methods are barcode-aware and can be customized via method parameters.
        - Aggregation to oligo-level data is supported for downstream analysis.
    """  # noqa: E501

    def _barcode_counts(self) -> NDArray[np.int32]:
        return (
            pd.DataFrame(
                self.observed * ~self.var_filter.T,
                index=self.obs_names,
                columns=self.var_names,
            )
            .T.groupby(self.oligos, observed=True)
            .transform("sum")
        ).T.values

    def drop_barcode_counts(self) -> None:

        self.LOGGER.info("Dropping barcode counts")

        self.data.layers.pop("barcode_counts", None)

    @property
    def oligo_data(self) -> "MPRAOligoData":
        """MPRAOligoData: Returns an instance of `MPRAOligoData` containing aggregated oligo-level data."""
        self.LOGGER.info("Computing oligo data")

        return self._oligo_data()

    @classmethod
    def from_file(cls, file_path: str) -> "MPRABarcodeData":

        data = pd.read_csv(file_path, sep="\t", header=0, index_col=0)
        data = data.fillna(0)

        replicate_columns_rna = data.columns[2::2]
        replicate_columns_dna = data.columns[1::2]

        anndata_replicate_rna = data[replicate_columns_rna].transpose().astype(np.int32)
        anndata_replicate_dna = data[replicate_columns_dna].transpose().astype(np.int32)

        anndata_replicate_rna.index = pd.Index([replicate.split("_")[2] for replicate in replicate_columns_rna])
        anndata_replicate_dna.index = pd.Index([replicate.split("_")[2] for replicate in replicate_columns_dna])

        adata = ad.AnnData(anndata_replicate_rna)
        adata.layers["rna"] = np.array(adata.X, dtype=np.int32)
        adata.layers["dna"] = np.asarray(anndata_replicate_dna.values, dtype=np.int32)

        adata.var["oligo"] = data["oligo_name"].astype("category")

        adata.uns["file_path"] = file_path
        adata.uns["date"] = pd.to_datetime("today").strftime("%Y-%m-%d")

        adata.uns["normalized"] = False
        adata.uns["barcode_threshold"] = None

        adata.varm["var_filter"] = pd.DataFrame(
            np.full((adata.n_vars, adata.n_obs), False),
            index=adata.var_names,
            columns=adata.obs_names,
        )

        return cls(adata)

    def complexity(self, method="lincoln") -> NDArray[np.int64]:
        """Calculates and returns the complexity of barcodes using the Lincoln-Peterson or Chapman estimation.

        Args:
            method (str): Either "lincoln" or "chapman".

        Returns:
           The Lincoln-Peterson or Chapman estimate.
        """

        if method not in {"lincoln", "chapman"}:
            raise ValueError("Method must be either 'lincoln' or 'chapman'.")

        n_observed = np.sum(self.observed, axis=1)
        num_rows = self.observed.shape[0]
        results = np.zeros((num_rows, num_rows), dtype=np.int64)
        for i in range(num_rows):
            for j in range(i, num_rows):
                n_recap = np.sum(np.sum(np.logical_and(self.observed[i, :], self.observed[j, :])))
                if method == "lincoln":
                    count = (n_observed[i] * n_observed[j]) / n_recap if n_recap > 0 else 0
                elif method == "chapman":
                    count = ((n_observed[i] + 1) * (n_observed[j] + 1) / (n_recap + 1)) - 1

                count = int(np.floor(count))  # type: ignore
                results[i, j] = count
                results[j, i] = count  # symmetric

        return results

    def _normalize(self) -> None:

        self.drop_normalized()

        self.LOGGER.info("Normalizing data")

        self.data.layers["dna_normalized"] = self._normalize_layer(
            self.dna_counts, self.observed, ~self.var_filter.T, self.scaling, self.pseudo_count
        )
        self.data.layers["rna_normalized"] = self._normalize_layer(
            self.rna_counts, self.observed, ~self.var_filter.T, self.scaling, self.pseudo_count
        )
        self._add_metadata("normalized", True)

    def _normalize_layer(
        self,
        counts: NDArray[np.int32],
        observed: NDArray[np.bool_],
        not_var_filter: NDArray[np.bool_],
        scaling: float,
        pseudocount: int,
    ) -> NDArray[np.float32]:

        # I do a pseudo count when normalizing to avoid division by zero when computing logfold ratios.
        # barcode filter has to be used again because we want to have a zero on filtered values.
        total_counts = np.sum((counts + (pseudocount * observed)) * not_var_filter, axis=1)

        # Avoid division by zero when pseudocount is set to 0
        total_counts[total_counts == 0] = 1
        return ((counts + (pseudocount * observed)) / total_counts[:, np.newaxis] * scaling) * not_var_filter

    def _oligo_data(self) -> "MPRAOligoData":

        # Convert the result back to an AnnData object
        oligo_data = ad.AnnData(self._sum_counts_by_oligo(self.rna_counts))

        oligo_data.layers["rna"] = np.array(oligo_data.X)
        oligo_data.layers["dna"] = self._sum_counts_by_oligo(self.dna_counts)

        oligo_data.layers["barcode_counts"] = (
            pd.DataFrame(
                self.barcode_counts,
                index=self.obs_names,
                columns=self.var_names,
            )
            .T.groupby(self.oligos, observed=True)
            .first()
            .T
        )

        oligo_data.obs_names = self.obs_names
        oligo_data.var_names = self.data.var["oligo"].unique().tolist()

        # Subset of vars using the firs occurence of oligo name
        indices = self.data.var["oligo"].drop_duplicates(keep="first").index
        oligo_data.var = self.data.var.loc[indices]

        oligo_data.obs = self.data.obs

        for key, value in self.data.uns.items():
            oligo_data.uns[f"MPRABarcodeData_{key}"] = value

        oligo_data.uns["SCALING"] = self.scaling
        oligo_data.uns["PSEUDOCOUNT"] = self.pseudo_count
        oligo_data.uns["correlation_activity"] = False
        oligo_data.uns["correlation_rna_normalized"] = False
        oligo_data.uns["correlation_dna_normalized"] = False

        return MPRAOligoData(oligo_data, self.barcode_threshold)

    def _sum_counts_by_oligo(self, counts: NDArray[np.int32]) -> pd.DataFrame:

        grouped = pd.DataFrame(
            counts,
            index=self.obs_names,
            columns=self.var_names,
        ).T.groupby(self.data.var["oligo"], observed=True)

        # Perform an operation on each group, e.g., mean
        return grouped.sum().T

    def _supporting_barcodes_per_oligo(self) -> pd.DataFrame:

        grouped = pd.DataFrame(
            self.observed * ~self.var_filter.T,
            index=self.obs_names,
            columns=self.var_names,
        ).T.groupby(self.oligos, observed=True)

        return grouped.apply(lambda x: x.sum()).T

    def _barcode_filter_rna_zscore(self, times_zscore=3):

        barcode_mask = pd.DataFrame(
            self.raw_dna_counts + self.raw_rna_counts,
            index=self.obs_names,
            columns=self.var_names,
        ).T.apply(lambda x: (x != 0))

        df_rna = pd.DataFrame(self.raw_rna_counts, index=self.obs_names, columns=self.var_names).T
        grouped = df_rna.where(barcode_mask).groupby(self.oligos, observed=True)

        mask = ((df_rna - grouped.transform("mean")) / grouped.transform("std")).abs() > times_zscore

        return mask

    def _barcode_filter_mad(self, times_mad=3, n_bins=20):

        # sum up DNA and RNA counts across replicates
        DNA_sum = pd.DataFrame(self.raw_dna_counts, index=self.obs_names, columns=self.var_names).T.sum(axis=1)
        RNA_sum = pd.DataFrame(self.raw_rna_counts, index=self.obs_names, columns=self.var_names).T.sum(axis=1)
        df_sums = pd.DataFrame({"DNA_sum": DNA_sum, "RNA_sum": RNA_sum}).fillna(0)
        # removing all barcodes with 0 counts in RNA an more DNA count than number of replicates/observations
        df_sums = df_sums[(df_sums["DNA_sum"] > self.data.n_obs) & (df_sums["RNA_sum"] > 0)]

        # remove all barcodes where oligo has less barcodes as the number of replicates/observations
        df_sums = df_sums.groupby(self.data.var["oligo"], observed=True).filter(lambda x: len(x) >= self.data.n_obs)

        # Calculate ratio, ratio_med, ratio_diff, and mad
        df_sums["ratio"] = np.log2(df_sums["DNA_sum"] / df_sums["RNA_sum"])
        df_sums["ratio_med"] = df_sums.groupby(self.data.var["oligo"], observed=True)["ratio"].transform("median")
        df_sums["ratio_diff"] = df_sums["ratio"] - df_sums["ratio_med"]
        # df_sums['mad'] = (df_sums['ratio'] - df_sums['ratio_med']).abs().mean()

        # Calculate quantiles within  n_bins
        qs = np.unique(np.quantile(np.log10(df_sums["RNA_sum"]), np.arange(0, n_bins) / n_bins))

        # Create bins based on rna_count
        df_sums["bin"] = pd.cut(
            np.log10(df_sums["RNA_sum"]),
            bins=qs,
            include_lowest=True,
            labels=[str(i) for i in range(0, len(qs) - 1)],
        )
        # Filter based on ratio_diff and mad
        df_sums["ratio_diff_med"] = df_sums.groupby("bin", observed=True)["ratio_diff"].transform("median")
        df_sums["ratio_diff_med_dist"] = np.abs(df_sums["ratio_diff"] - df_sums["ratio_diff_med"])
        df_sums["mad"] = df_sums.groupby("bin", observed=True)["ratio_diff_med_dist"].transform("median")

        m = df_sums.ratio_diff > times_mad * df_sums.mad
        df_sums = df_sums[~m]

        return self.var_filter.apply(lambda col: col | ~self.var_filter.index.isin(df_sums.index))

    def _barcode_filter_random(
        self, proportion: float = 1.0, total: Optional[int] = None, aggegate_over_replicates: bool = True
    ):

        if aggegate_over_replicates and total is None:
            total = self.var_filter.shape[0]
        elif total is None:
            total = self.var_filter.size

        num_true_cells = int(total * (1.0 - proportion))  # type: ignore
        true_indices = np.random.choice(total, num_true_cells, replace=False)

        mask = pd.DataFrame(
            np.full((self.data.n_vars, self.data.n_obs), False),
            index=self.var_names,
            columns=self.obs_names,
        )

        if aggegate_over_replicates:
            mask.iloc[true_indices, :] = True
        else:
            flat_df = mask.values.flatten()

            flat_df[true_indices] = True

            mask = pd.DataFrame(
                flat_df.reshape(self.var_filter.shape),
                index=self.var_names,
                columns=self.obs_names,
            )
        return mask

    def _barcode_filter_min_count(self, rna_min_count=None, dna_min_count=None):

        return self._barcode_filter_min_max_count(BarcodeFilter.MIN_COUNT, rna_min_count, dna_min_count)

    def _barcode_filter_max_count(self, rna_max_count=None, dna_max_count=None):

        return self._barcode_filter_min_max_count(BarcodeFilter.MAX_COUNT, rna_max_count, dna_max_count)

    def _barcode_filter_min_max_counts(self, barcode_filter, counts, count_threshold):
        if barcode_filter == BarcodeFilter.MIN_COUNT:
            return (counts < count_threshold).T
        elif barcode_filter == BarcodeFilter.MAX_COUNT:
            return (counts > count_threshold).T

    def _barcode_filter_min_max_count(self, barcode_filter, rna_count=None, dna_count=None):
        mask = pd.DataFrame(
            np.full((self.n_vars, self.n_obs), False),
            index=self.var_names,
            columns=self.obs_names,
        )
        if rna_count is not None:
            mask = mask | self._barcode_filter_min_max_counts(barcode_filter, self.raw_rna_counts, rna_count)
        if dna_count is not None:
            mask = mask | self._barcode_filter_min_max_counts(barcode_filter, self.raw_dna_counts, dna_count)

        return mask

    def apply_barcode_filter(self, barcode_filter: BarcodeFilter, params: dict = {}) -> None:
        """Applies a specified barcode filter to the dataset using the provided parameters.

        This method selects the appropriate barcode filtering function based on the `barcode_filter` argument and applies it to update the `var_filter` attribute. Supported filters include RNA z-score, MAD, random, minimum count, and maximum count. After applying the filter, metadata is updated to record the applied filter.

        Args:
            barcode_filter (BarcodeFilter): The type of barcode filter to apply.
            params (dict, optional): Additional parameters to pass to the filter function. Defaults to an empty dictionary.

        Raises:
            ValueError: If an unsupported barcode filter is provided.
        """  # noqa: E501

        filter_switch = {
            BarcodeFilter.RNA_ZSCORE: self._barcode_filter_rna_zscore,
            BarcodeFilter.MAD: self._barcode_filter_mad,
            BarcodeFilter.RANDOM: self._barcode_filter_random,
            BarcodeFilter.MIN_COUNT: self._barcode_filter_min_count,
            BarcodeFilter.MAX_COUNT: self._barcode_filter_max_count,
        }

        filter_func = filter_switch.get(barcode_filter)
        if filter_func:
            self.var_filter = self.var_filter | filter_func(**params)
        else:
            raise ValueError(f"Unsupported barcode filter: {barcode_filter}")

        self._add_metadata("var_filter", [barcode_filter.value])

    def drop_count_sampling(self) -> None:
        """Removes count sampling data from the dataset.

        This method performs the following actions:
        - Calls `drop_normalized()` to remove any normalized data.
        - Logs the action of dropping count sampling.
        - Deletes the "count_sampling" entry from the `.uns` attribute of the data.
        - Removes "rna_sampling" and "dna_sampling" layers from the data, if they exist.
        """

        self.drop_normalized()

        self.LOGGER.info("Dropping count sampling")

        del self.data.uns["count_sampling"]
        self.data.layers.pop("rna_sampling", None)
        self.data.layers.pop("dna_sampling", None)

    def _calculate_proportions(
        self,
        proportion: Optional[float],
        total: Optional[int],
        aggregate_over_replicates: bool,
        counts: NDArray[np.int32],
        replicates: int,
    ):
        pp = [1.0] * replicates

        if proportion is not None:
            pp = [proportion] * replicates

        if total is not None:
            if aggregate_over_replicates:
                for i, p in enumerate(pp):
                    pp[i] = min(float(total) / float(np.sum(counts)), p)
            else:
                for i, p in enumerate(pp):
                    pp[i] = min(float(total) / float(np.sum(counts[i, :])), p)
        return pp

    def _sample_individual_counts(self, x, proportion: float) -> int:
        return int(
            np.floor(x * proportion)
            + (0.0 if x != 0 or np.random.rand() > (x * proportion - np.floor(x * proportion)) else 1.0)
        )

    def _apply_sampling(
        self,
        layer_name: str,
        counts: NDArray[np.int32],
        proportion: Optional[float],
        total: Optional[int],
        max_value: Optional[int],
        aggregate_over_replicates: bool,
    ) -> None:

        sampled_counts = counts.copy()

        if total is not None or proportion is not None:

            pp = self._calculate_proportions(proportion, total, aggregate_over_replicates, sampled_counts, self.n_obs)

            vectorized_sample_individual_counts = np.vectorize(self._sample_individual_counts)

            for i, p in enumerate(pp):
                sampled_counts[i, :] = vectorized_sample_individual_counts(sampled_counts[i, :], proportion=p)

        if max_value is not None:
            sampled_counts = np.clip(sampled_counts, None, max_value)

        self.data.layers[layer_name] = sampled_counts

    def apply_count_sampling(
        self,
        count_type: CountSampling,
        proportion: Optional[float] = None,
        total: Optional[int] = None,
        max_value: Optional[int] = None,
        aggregate_over_replicates: bool = False,
    ) -> None:
        """Applies count sampling to RNA and/or DNA count data according to the specified parameters.

        Args:
            count_type (CountSampling): Specifies which counts to sample. Options are RNA, DNA, or RNA_AND_DNA.
            proportion (Optional[float]): Proportion of counts to sample (between 0 and 1). If None, this parameter is ignored.
            total (Optional[int]): Total number of counts to sample. If None, this parameter is ignored.
            max_value (Optional[int]): Maximum value for sampled counts. If None, this parameter is ignored.
            aggregate_over_replicates (bool): Whether to aggregate counts over replicates before sampling.

        Side Effects:
            - Adds sampling metadata to the object.
            - Drops any normalized data associated with the object.
        """

        if count_type == CountSampling.RNA or count_type == CountSampling.RNA_AND_DNA:
            self._apply_sampling("rna_sampling", self.rna_counts, proportion, total, max_value, aggregate_over_replicates)

        if count_type == CountSampling.DNA or count_type == CountSampling.RNA_AND_DNA:
            self._apply_sampling("dna_sampling", self.dna_counts, proportion, total, max_value, aggregate_over_replicates)

        self._add_metadata(
            "count_sampling",
            [
                {
                    count_type.value: {
                        "proportion": proportion,
                        "total": total,
                        "max_value": max_value,
                        "aggregate_over_replicates": aggregate_over_replicates,
                    }
                }
            ],
        )

        self.drop_normalized()
        self.drop_barcode_counts()


class MPRAOligoData(MPRAData):
    """MPRAOligoData is a subclass of MPRAData designed to handle MPRA (Massively Parallel Reporter Assay) oligo-level data.

    This class provides methods for loading, normalizing, and managing barcode counts and associated data layers for MPRA experiments. Barcode counts must be pre-set before accessing, as they cannot be computed within this class. The normalization process includes pseudocount handling to avoid division by zero and supports per-barcode normalization.

    Raises:
        MPRAlibException: If barcode counts are not set when accessed.
    """  # noqa: E501

    def _barcode_counts(self):
        raise MPRAlibException(
            "Barcode counts are not set in MPRAOligoData and cannot be computed. They have to be pre-set before accessing."
        )

    def drop_barcode_counts(self):
        pass

    @classmethod
    def from_file(cls, file_path: str) -> "MPRAOligoData":

        return MPRAOligoData(ad.read(file_path))

    def _normalize(self) -> None:

        self.drop_normalized()

        self.LOGGER.info("Normalizing data")

        self.data.layers["dna_normalized"] = self._normalize_layer(
            self.dna_counts, ~self.var_filter.T, self.barcode_counts, self.scaling, self.pseudo_count
        )
        self.data.layers["rna_normalized"] = self._normalize_layer(
            self.rna_counts, ~self.var_filter.T, self.barcode_counts, self.scaling, self.pseudo_count
        )
        self._add_metadata("normalized", True)

    def _normalize_layer(
        self,
        counts: NDArray[np.int32],
        not_var_filter: NDArray[np.bool_],
        barcode_counts: NDArray[np.int32],
        scaling: float,
        pseudocount: int,
    ) -> NDArray[np.float32]:

        # I do a pseudo count when normalizing to avoid division by zero when computing logfold ratios.
        # Pseudocount has also be done per barcode.
        # var filter has to be used again because we want to have a zero on filtered values.
        total_counts = np.sum((counts + (pseudocount * barcode_counts)) * not_var_filter, axis=1)

        # Avoid division by zero when pseudocount is set to 0
        total_counts[total_counts == 0] = 1
        scaled_counts = (counts + (pseudocount * barcode_counts)) / total_counts[:, np.newaxis] * scaling
        return (
            np.divide(scaled_counts, barcode_counts, where=barcode_counts != 0, out=np.zeros_like(scaled_counts))
            * not_var_filter
        )



class MPRAlibException(Exception):
    """MPRAlib error class for specific exceptions.

    Args:
        message (str): A description of the error.
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


class IOException(MPRAlibException):
    """Exception raised for IO-related errors.

    Args:
        message (str): A description of the IO error.
    """

    def __init__(self, message: str):
        super().__init__(message)


class SequenceDesignException(IOException):
    """Exception raised for errors related to sequence design file.

    Args:
        message (str): A description of the sequence design error.
    """

    def __init__(self, column: str, file_path: str):
        super().__init__(f"Column {column} has a wrong format in file {file_path}.")



from enum import Enum
import json
from importlib.resources import files
import jsonschema
import csv
import gzip
import ast
from mpralib.utils.io import is_compressed_file
import tqdm
import logging
import re
from typing import Optional


class ValidationSchema(Enum):
    REPORTER_SEQUENCE_DESIGN = "reporter_sequence_design"
    REPORTER_BARCODE_TO_ELEMENT_MAPPING = "reporter_barcode_to_element_mapping"
    REPORTER_EXPERIMENT_BARCODE = "reporter_experiment_barcode"
    REPORTER_EXPERIMENT = "reporter_experiment"
    REPORTER_ELEMENT = "reporter_element"
    REPORTER_VARIANT = "reporter_variant"
    REPORTER_GENOMIC_ELEMENT = "reporter_genomic_element"
    REPORTER_GENOMIC_VARIANT = "reporter_genomic_variant"


class SchemaToFileNameMap:
    def __init__(self):
        self._data = {}

    def set(self, key: ValidationSchema, file_name: str):
        if not isinstance(key, ValidationSchema):
            raise KeyError(f"Key must be a FileKey enum value. Got {key}")
        if not isinstance(file_name, str):
            raise ValueError("File path must be a string")
        self._data[key] = file_name

    def get(self, key: ValidationSchema):
        return self._data.get(key, None)

    def as_dict(self):
        return {k.value: v for k, v in self._data.items()}


schemaFilemap = SchemaToFileNameMap()
schemaFilemap.set(ValidationSchema.REPORTER_SEQUENCE_DESIGN, "reporter_sequence_design.json")
schemaFilemap.set(ValidationSchema.REPORTER_BARCODE_TO_ELEMENT_MAPPING, "reporter_barcode_to_element_mapping.json")
schemaFilemap.set(ValidationSchema.REPORTER_EXPERIMENT_BARCODE, "reporter_experiment_barcode.json")
schemaFilemap.set(ValidationSchema.REPORTER_EXPERIMENT, "reporter_experiment.json")
schemaFilemap.set(ValidationSchema.REPORTER_ELEMENT, "reporter_element.json")
schemaFilemap.set(ValidationSchema.REPORTER_VARIANT, "reporter_variant.json")
schemaFilemap.set(ValidationSchema.REPORTER_GENOMIC_ELEMENT, "reporter_genomic_element.json")
schemaFilemap.set(ValidationSchema.REPORTER_GENOMIC_VARIANT, "reporter_genomic_variant.json")


def _convert_row_value(value: str, prop_schema: dict):
    try:
        if prop_schema.get("type") == "integer":
            converted_value = int(value)
        elif prop_schema.get("type") == "number":
            converted_value = float(value)
        elif prop_schema.get("type") == "array":
            converted_value = ast.literal_eval(value)
        else:
            converted_value = value
    except ValueError:
        converted_value = value  # Let validation catch the error

    return converted_value


def validate_tsv_with_schema(tsv_file_path: str, schema_type: ValidationSchema) -> bool:
    """Validates a TSV file against a specified JSON schema.

    This function reads a TSV file (optionally gzipped), converts each row to a dictionary,
    and validates each row against the provided JSON schema. If any row fails validation,
    a warning is logged. If an unexpected error occurs during validation, it is logged and raised.

    Args:
        tsv_file_path (str): Path to the TSV file to validate. The file may be gzipped.
        schema_type (ValidationSchema): The type of schema to validate against.

    Returns:
        True if all rows are valid according to the schema, False otherwise.

    Raises:
        Exception: If an unexpected error occurs during validation.

    Logs:
        - Warnings for each row that fails schema validation.
        - Errors for unexpected exceptions during validation.
        - Info if the file is valid according to the schema.
        - Warning if the file is not valid according to the schema.
    """
    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.WARNING)

    schema = _load_schema(schema_type)
    header = _get_header_for_schema(schema_type)
    open_func = gzip.open if is_compressed_file(tsv_file_path) else open
    mode = "rt" if is_compressed_file(tsv_file_path) else "r"

    correct_file = True

    with open_func(tsv_file_path, mode, encoding="utf-8") as tsvfile:
        reader = csv.DictReader(tsvfile, delimiter="\t", fieldnames=header)  # type: ignore

        for i, row in enumerate(tqdm.tqdm(reader, desc="Validating rows", unit="row"), start=1):
            _convert_row_types(row, schema)
            try:
                jsonschema.validate(instance=row, schema=schema)
            except jsonschema.ValidationError as e:
                LOGGER.warning(f"Row {i} invalid: {e.message}")
                correct_file = False
            except Exception as e:
                LOGGER.error(f"Row {i} error: {e}")
                correct_file = False
                raise e

    if correct_file:
        LOGGER.info(f"File {tsv_file_path} is valid according to schema {schema_type.value}.")
    else:
        LOGGER.warning(f"File {tsv_file_path} is not valid according to schema {schema_type.value}.")
    return correct_file


def _load_schema(schema_type: ValidationSchema):
    schema_filename = schemaFilemap.get(schema_type)
    if schema_filename is None:
        raise ValueError(f"No schema file mapped for schema type: {schema_type}")
    schema_path = files("mpralib.schemas").joinpath(schema_filename)
    with schema_path.open("r", encoding="utf-8") as f:
        return json.load(f)


def _get_header_for_schema(schema_type: ValidationSchema) -> Optional[list]:
    if schema_type == ValidationSchema.REPORTER_BARCODE_TO_ELEMENT_MAPPING:
        return ["barcode", "oligoName"]
    elif schema_type == ValidationSchema.REPORTER_GENOMIC_ELEMENT:
        return [
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "log2FoldChange",
            "inputCount",
            "outputCount",
            "minusLog10PValue",
            "minusLog10QValue",
        ]
    elif schema_type == ValidationSchema.REPORTER_GENOMIC_VARIANT:
        return [
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "log2FoldChange",
            "inputCountRef",
            "outputCountRef",
            "inputCountAlt",
            "outputCountAlt",
            "minusLog10PValue",
            "minusLog10QValue",
            "postProbEffect",
            "CI_lower_95",
            "CI_upper_95",
            "variantPos",
            "refAllele",
            "altAllele",
        ]
    return None


def _convert_row_types(row: dict, schema: dict) -> None:
    # Handle patternProperties
    for prop_pattern_string, prop_schema in schema.get("patternProperties", {}).items():
        prop_pattern = re.compile(prop_pattern_string)
        for prop in [p for p in row if prop_pattern.match(p)]:
            if row[prop] != "":
                if "anyOf" in prop_schema:
                    for anyOfProp_schema in prop_schema["anyOf"]:
                        row[prop] = _convert_row_value(row[prop], anyOfProp_schema)
                else:
                    row[prop] = _convert_row_value(row[prop], prop_schema)
    # Handle properties
    for prop, prop_schema in schema.get("properties", {}).items():
        if prop in row and row[prop] != "":
            if "anyOf" in prop_schema:
                for anyOfProp_schema in prop_schema["anyOf"]:
                    row[prop] = _convert_row_value(row[prop], anyOfProp_schema)
            else:
                row[prop] = _convert_row_value(row[prop], prop_schema)



import pandas as pd
import numpy as np
import ast
from importlib.resources import files
from mpralib.mpradata import MPRABarcodeData, MPRAOligoData, MPRAData
from mpralib.exception import SequenceDesignException, MPRAlibException
from typing import Optional


def chromosome_map() -> pd.DataFrame:

    with files("mpralib.data").joinpath("hg19.chromAlias.txt").open() as chromAlias_hg19:
        df = pd.read_csv(chromAlias_hg19, sep="\t", header=None, comment="#", dtype="category")
    df["release"] = "GRCh37"

    with files("mpralib.data").joinpath("hg38.chromAlias.txt").open() as chromAlias_hg38:
        df_38 = pd.read_csv(chromAlias_hg38, sep="\t", header=None, comment="#", dtype="category")
    df_38["release"] = "GRCh38"

    df = pd.concat([df, df_38], ignore_index=True)
    df.columns = ["ucsc", "assembly", "genbank", "refseq", "release"]
    return df


def is_compressed_file(filepath):
    """Check if a file is compressed (gzip or bgz).

    Args:
        filepath (str): Path to the file to check.

    Returns:
        True if the file is compressed, False otherwise.
    """
    return is_gzip_file(filepath) or is_bgzf(filepath)


def is_gzip_file(filepath):
    """Check if a file is a gzip-compressed file based on its magic number.

    Args:
        filepath (str or Path): Path to the file to check.

    Returns:
        True if the file is gzip-compressed, False otherwise.
    """
    with open(filepath, "rb") as f:
        magic = f.read(2)
    return magic == b"\x1f\x8b"


def is_bgzf(filepath):
    """Check if a file is in BGZF (Blocked GNU Zip Format) format.

    BGZF is a variant of the standard gzip format with extra fields that allow for random access.
    This function reads the file header and checks for the BGZF-specific magic numbers and flags.

    Args:
        filepath (str): Path to the file to be checked.

    Returns:
        True if the file is in BGZF format, False otherwise.
    """
    with open(filepath, "rb") as f:
        header = f.read(18)
    return (
        len(header) >= 18
        and header[0:2] == b"\x1f\x8b"  # gzip magic number
        and header[3] == 4  # FLG.FEXTRA set
        and header[12:14] == b"BC"  # BGZF extra subfield
    )


def export_activity_file(mpradata: MPRAOligoData, output_file_path: str) -> None:
    """Export activity data from an MPRAdata object to a tab-separated values (TSV) file.

    The function processes the grouped data from the MPRAdata object, extracts relevant information for each replicate, and writes the data to a TSV file. The output file contains columns for replicate, oligo name, DNA counts, RNA counts, normalized DNA counts, normalized RNA counts, log2 fold change, and the number of barcodes. Barcode filters, count sampling and barcode thresholds are applied.

    Args:
        mpradata (MPRAdata): An object containing MPRA (Massively Parallel Reporter Assay) data.
        output_file_path (str): The file path where the output TSV file will be saved.
    """  # noqa: E501

    output = pd.DataFrame()

    mpradata.activity

    for replicate in mpradata.obs_names:
        replicate_data = mpradata.data[replicate, :]
        replicate_data = replicate_data[
            :, np.asarray(replicate_data.layers["barcode_counts"]) >= np.asarray(mpradata.barcode_threshold)
        ]
        df = {
            "replicate": np.repeat(replicate, replicate_data.var_names.size),
            "oligo_name": replicate_data.var["oligo"],
            "dna_counts": np.asarray(replicate_data.layers["dna"])[0, :],
            "rna_counts": np.asarray(replicate_data.layers["rna"])[0, :],
            "dna_normalized": np.round(np.asarray(replicate_data.layers["dna_normalized"])[0, :], 4),
            "rna_normalized": np.round(np.asarray(replicate_data.layers["rna_normalized"])[0, :], 4),
            "log2FoldChange": np.round(np.asarray(replicate_data.layers["activity"])[0, :], 4),
            "n_bc": np.asarray(replicate_data.layers["barcode_counts"])[0, :],
        }
        output = pd.concat([output, pd.DataFrame(df)], axis=0)

    output.to_csv(output_file_path, sep="\t", index=False)


def export_barcode_file(mpradata: MPRABarcodeData, output_file_path: str) -> None:
    """Export barcode count data to a file.

    This function takes an MPRAdata object and exports its barcode count data to a specified file path in tab-separated values (TSV) format. The output file will contain columns for barcodes, oligo names, and DNA/RNA counts for each replicate. Modifides counts (barcode filter/sampling) if applicable will be written.

    Args:
        mpradata (MPRAdata): An object containing MPRA data, including barcodes, oligos, DNA counts, RNA counts, and replicates.
        output_file_path (str): The file path where the output TSV file will be saved.
    """  # noqa: E501

    output = pd.DataFrame({"barcode": mpradata.var_names, "oligo_name": mpradata.oligos})

    mpradata.normalized_rna_counts

    dna_counts = mpradata.dna_counts
    rna_counts = mpradata.rna_counts
    for i, replicate in enumerate(mpradata.obs_names):
        output[f"dna_count_{replicate}"] = dna_counts[i]
        output[f"rna_count_{replicate}"] = rna_counts[i]
    output.replace(0, "", inplace=True)
    output.to_csv(output_file_path, sep="\t", index=False)


def export_counts_file(
    mpradata: MPRAData, output_file_path: str, normalized: bool = False, filter: Optional[np.ndarray] = None
) -> None:
    if isinstance(mpradata, MPRABarcodeData):
        df = {"ID": mpradata.var_names}
    elif isinstance(mpradata, MPRAOligoData):
        df = {"ID": mpradata.oligos}
    else:
        raise MPRAlibException(f"Invalid MPRA data type {type(mpradata)}. Expected MPRAOligoData or MPRABarcodeData.")

    if normalized:
        dna_counts = mpradata.normalized_dna_counts.copy()
        rna_counts = mpradata.normalized_rna_counts.copy()
    else:
        dna_counts = mpradata.dna_counts.copy()
        rna_counts = mpradata.rna_counts.copy()

    not_observed_mask = ~mpradata.observed
    lower_bc_mask = mpradata.barcode_counts < mpradata.barcode_threshold
    mask = [not_observed_mask, lower_bc_mask]
    if filter is not None:
        mask.append(filter)

    dna_counts = np.ma.masked_array(dna_counts, mask=np.any(mask, axis=0))
    rna_counts = np.ma.masked_array(rna_counts, mask=np.any(mask, axis=0))

    for idx, replicate in enumerate(mpradata.obs_names):
        df["dna_count_" + replicate] = dna_counts[idx, :]
        df["rna_count_" + replicate] = rna_counts[idx, :]

    df = pd.DataFrame(df).set_index("ID")
    df = df.map(lambda x: np.nan if isinstance(x, np.ma.core.MaskedConstant) else x)
    nan_columns = df.isna().all(axis=1)

    if isinstance(mpradata, MPRABarcodeData):
        df.insert(0, "oligo_name", mpradata.oligos)

    # remove IDs which are all zero
    df = df[~nan_columns]

    if normalized:
        df.to_csv(output_file_path, sep="\t", index=True, na_rep="", float_format="%.6f")
    else:
        df.to_csv(output_file_path, sep="\t", index=True, na_rep="", float_format="%.0f")


def read_sequence_design_file(file_path: str) -> pd.DataFrame:
    """Read sequence design from a tab-separated values (TSV) file.

    This function reads metadata from a TSV file and returns it as a pandas DataFrame. The metadata file should contain columns for sample ID, replicate, and any additional metadata. The sample ID should correspond to the oligo name in the MPRA data object.

    Args:
        file_path (str): The file path of the metadata TSV file.

    Returns:
        A DataFrame containing the metadata.
    """  # noqa: E501

    df = pd.read_csv(
        file_path,
        sep="\t",
        header=0,
        na_values=["NA"],
        usecols=[
            "name",
            "sequence",
            "category",
            "class",
            "source",
            "ref",
            "chr",
            "start",
            "end",
            "strand",
            "variant_class",
            "variant_pos",
            "SPDI",
            "allele",
            "info",
        ],
        dtype={
            "name": "category",
            "sequence": str,
            "category": "category",
            "class": "category",
            "source": str,
            "ref": str,
            "chr": "category",
            "start": "Int64",
            "end": "Int64",
            "strand": "category",
            "variant_class": str,
            "variant_pos": str,
            "SPDI": str,
            "allele": str,
            "info": str,
        },
    ).drop_duplicates()

    # Set specific columns as arrays
    df["variant_class"] = df["variant_class"].fillna("[]").apply(ast.literal_eval)
    df["variant_pos"] = df["variant_pos"].fillna("[]").apply(ast.literal_eval)
    df["SPDI"] = df["SPDI"].fillna("[]").apply(ast.literal_eval)
    df["allele"] = df["allele"].fillna("[]").apply(ast.literal_eval)

    # Set specific columns as categorical or integer types
    df["category"] = pd.Categorical(df["category"])
    df["class"] = pd.Categorical(df["class"])
    df["chr"] = pd.Categorical(df["chr"])
    df["strand"] = pd.Categorical(df["strand"])
    df["name"] = pd.Categorical(df["name"].str.replace(r"[\s\[\]]", "_", regex=True))

    # oligo name as index
    df.set_index("name", inplace=True)

    # Validate that the 'sequence' column contains only valid DNA characters
    if np.any(~df["sequence"].str.match(r"^[ATGCatgc]+$", na=False)):
        raise SequenceDesignException("sequence", file_path)

    # Validate that the 'class' column contains only 'variant', 'element', 'synthetic' or 'scrambled'
    valid_categories = {"variant", "element", "synthetic", "scrambled"}
    if not set(df["category"].cat.categories).issubset(valid_categories):
        raise SequenceDesignException("category", file_path)

    # Validate that the 'category' column contains only:
    valid_classes = {
        "test",
        "variant positive control",
        "variant negative control",
        "element active control",
        "element inactive control",
    }
    if not set(df["class"].cat.categories).issubset(valid_classes):
        raise SequenceDesignException("class", file_path)

    return df



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from mpralib.mpradata import MPRAData, Modality, MPRAOligoData, MPRABarcodeData

custom_params = {"axes.spines.right": False, "axes.spines.top": False}
custom_palette = sns.color_palette(["#72ACBF", "#BF2675", "#2ecc71", "#f1c40f", "#9b59b6"])
sns.set_theme(style="white", rc=custom_params, palette=custom_palette)


def correlation(data: MPRAData, layer: Modality, replicates=None) -> sns.PairGrid:

    counts = None
    if layer == Modality.DNA:
        counts = data.dna_counts.copy()
    elif layer == Modality.RNA:
        counts = data.rna_counts.copy()
    elif layer == Modality.RNA_NORMALIZED:
        counts = data.normalized_rna_counts.copy()
    elif layer == Modality.DNA_NORMALIZED:
        counts = data.normalized_dna_counts.copy()
    elif layer == Modality.ACTIVITY:
        counts = data.activity.copy()

    counts = np.ma.masked_array(counts, mask=[data.barcode_counts < data.barcode_threshold])

    if replicates:
        idx = np.array([data.obs_names.get_loc(rep) for rep in replicates])
        counts = pd.DataFrame(counts[idx].T, columns=[f"Replicate {i}" for i in data.obs_names[idx]], index=data.var_names)
    else:
        counts = pd.DataFrame(counts.T, columns=[f"Replicate {i}" for i in data.obs_names], index=data.var_names)

    g = sns.PairGrid(counts)
    g.map_upper(sns.scatterplot, s=1)
    g.map_diag(sns.kdeplot)
    g.map_lower(sns.kdeplot, fill=True)
    g.figure.suptitle("Correlation Plot")
    return g


def dna_vs_rna(data: MPRAData, replicates=None) -> sns.JointGrid:

    counts_dna = data.normalized_dna_counts.copy()
    counts_rna = data.normalized_rna_counts.copy()

    mask = [data.barcode_counts < data.barcode_threshold]

    counts_dna = np.ma.masked_array(counts_dna, mask=mask)
    counts_rna = np.ma.masked_array(counts_rna, mask=mask)

    if replicates:
        idx = [data.obs_names.get_loc(rep) for rep in replicates]
        counts_dna = counts_dna[idx]
        counts_rna = counts_rna[idx]

    median_dna = np.median(counts_dna, axis=0)
    median_rna = np.median(counts_rna, axis=0)

    median_dna = np.ma.masked_equal(median_dna, 0)
    median_rna = np.ma.masked_equal(median_rna, 0)

    df = pd.DataFrame({"DNA [log10]": np.log10(median_dna), "RNA [log10]": np.log10(median_rna)})

    g = sns.jointplot(data=df, x="DNA [log10]", y="RNA [log10]", kind="scatter", s=3)
    g.ax_joint.plot(
        [df["DNA [log10]"].min(), df["DNA [log10]"].max()],
        [df["RNA [log10]"].min(), df["RNA [log10]"].max()],
        linestyle="--",
        color="#BF2675",
    )
    g.figure.suptitle("Median normalized counts across replicates")
    g.figure.subplots_adjust(top=0.95)
    return g


def barcodes_per_oligo(data: MPRAOligoData, replicates=None) -> sns.FacetGrid:

    bc_counts = data.barcode_counts.copy()

    intercept_median = np.median(bc_counts, axis=1)
    intercept_mean = np.mean(bc_counts, axis=1)

    if replicates:
        idx = np.array([data.obs_names.get_loc(rep) for rep in replicates])
        bc_counts = pd.DataFrame(
            bc_counts[idx].T, columns=[f"Replicate {i}" for i in data.obs_names[idx]], index=data.var_names
        )
    else:
        replicates = data.obs_names
        bc_counts = pd.DataFrame(bc_counts.T, columns=[f"Replicate {i}" for i in data.obs_names], index=data.var_names)

    bc_counts = bc_counts.melt(var_name="replicate", value_name="n_bc")

    g = sns.FacetGrid(bc_counts, col="replicate")

    # Histogram plot
    g.map(sns.histplot, "n_bc")

    i = 0
    for ax in g.axes.flatten():
        ax.axvline(x=intercept_median[i], color="red", linestyle="--")
        ax.axvline(x=intercept_mean[i], color="blue", linestyle="--")
        ax.set_title(f"Replicate {replicates[i]}")
        ax.text(ax.get_xlim()[1] * 0.5, ax.get_ylim()[1] * 0.9, f"{intercept_median[i]:.0f}", color="red", ha="left")
        ax.text(ax.get_xlim()[1] * 0.5, ax.get_ylim()[1] * 0.8, f"{intercept_mean[i]:.2f}", color="blue", ha="left")
        i = i + 1
    g.set_axis_labels("Frequency", "Barcodes per oligo")

    return g


def barcodes_outlier(data: MPRABarcodeData) -> None:

    # counts_dna = data.normalized_dna_counts.copy()
    # counts_rna = data.normalized_rna_counts.copy()
    counts_dna = data.dna_counts.copy()
    counts_rna = data.rna_counts.copy()

    counts_dna_sum = counts_dna.sum(axis=0)
    counts_rna_sum = counts_rna.sum(axis=0)

    mask = np.any(
        [
            np.any(data.barcode_counts < data.barcode_threshold, axis=0),
            np.any([counts_dna_sum <= 10, counts_rna_sum <= 0], axis=0),
            ~np.any(data.observed, axis=0),
        ],
        axis=0,
    )

    print(np.shape(mask))
    print(np.sum(mask))

    counts_dna_sum = np.ma.masked_array(counts_dna_sum, mask=mask)
    counts_rna_sum = np.ma.masked_array(counts_rna_sum, mask=mask)

    counts_ratio = np.log2(counts_rna_sum / counts_dna_sum)
    counts = pd.DataFrame(
        {
            "dna": counts_dna_sum,
            "ratio": counts_ratio,
            "oligo": data.oligos,
        },
    )

    counts = counts[~mask]

    print(len(counts))

    counts["ratio_med"] = counts.groupby("oligo")["ratio"].transform("median")
    counts["ratio_diff"] = counts["ratio"] - counts["ratio_med"]

    nbin = 20
    qs = np.quantile(np.log10(counts["dna"]), np.arange(0, nbin + 1) / nbin)
    counts["bin"] = pd.cut(
        np.log10(counts["dna"]), bins=qs, include_lowest=True, labels=[str(i) for i in range(0, len(qs) - 1)]
    )

    stats = counts.groupby("bin").agg(mean_diff=("ratio_diff", "mean"), sd_diff=("ratio_diff", "std")).reset_index()

    # Get the last n categories of ordered categories
    # Plotting
    plt.figure(figsize=(10, 6))
    sampled_counts = counts.sample(n=min(10000, len(counts)))
    if isinstance(sampled_counts, pd.Series):
        sampled_counts = sampled_counts.to_frame().T
    sns.scatterplot(data=sampled_counts, x="bin", y="ratio_diff", alpha=0.3)
    filtered_counts = counts[(counts["ratio_diff"] > 5) & (counts["bin"].isin(counts["bin"].cat.categories[9:]))]
    if isinstance(filtered_counts, pd.Series):
        filtered_counts = filtered_counts.to_frame().T
    sns.scatterplot(
        data=filtered_counts,
        x="bin",
        y="ratio_diff",
        alpha=1,
        color="red",
    )
    plt.errorbar(x=stats["bin"], y=stats["mean_diff"], yerr=2 * stats["sd_diff"], fmt="o", color="dodgerblue", linewidth=1)
    plt.axhline(y=5, color="red")
    plt.xlabel("Bin")
    plt.ylabel("Ratio Difference")
    plt.title("Ratio Difference by Bin")
    plt.show()



import numpy as np
import pandas as pd
import anndata as ad
import copy
import pytest
from mpralib.mpradata import MPRABarcodeData, CountSampling, BarcodeFilter, Modality, MPRAOligoData


OBS = pd.DataFrame(index=["rep1", "rep2", "rep3"])
VAR = pd.DataFrame(
    {"oligo": ["oligo1", "oligo1", "oligo2", "oligo3", "oligo3"]},
    index=["barcode1", "barcode2", "barcode3", "barcode4", "barcode5"],
)
COUNTS_DNA = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
COUNTS_RNA = np.array([[1, 2, 4, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])

FILTER = np.array(
    [
        [False, True, False],
        [False, False, False],
        [True, False, False],
        [False, False, True],
        [False, False, True],
    ]
)


@pytest.fixture
def mpra_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


@pytest.fixture
def mpra_data_with_bc_filter(mpra_data):
    data = copy.deepcopy(mpra_data)
    data.var_filter = FILTER
    return data


def test_apply_count_sampling_rna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, proportion=0.5)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.all(rna_sampling <= np.asarray(mpra_data.data.layers["rna"]))
    assert np.all(rna_sampling >= 0)


def test_apply_count_sampling_dna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.DNA, proportion=0.5)
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(dna_sampling <= mpra_data.data.layers["dna"])
    assert np.all(dna_sampling >= 0)


def test_apply_count_sampling_rna_and_dna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, proportion=0.5)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(rna_sampling <= mpra_data.data.layers["rna"])
    assert np.all(rna_sampling >= 0)
    assert np.all(dna_sampling <= np.asarray(mpra_data.data.layers["dna"]))
    assert np.all(dna_sampling >= 0)


def test_apply_count_sampling_rna_total(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, total=10)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.sum(rna_sampling) <= 30


def test_apply_count_sampling_total(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, total=10)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.sum(rna_sampling) <= 30
    assert np.sum(dna_sampling) <= 30


def test_apply_count_sampling_max_value_rna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, max_value=2)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.all(rna_sampling <= 2)


def test_apply_count_sampling_max_value(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, max_value=2)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(rna_sampling <= 2)
    assert np.all(dna_sampling <= 2)


def test_apply_count_sampling_aggregate_over_replicates(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, total=10, aggregate_over_replicates=True)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.sum(rna_sampling) <= 11
    assert np.sum(dna_sampling) <= 11


def test_compute_supporting_barcodes(mpra_data):
    supporting_barcodes = mpra_data._supporting_barcodes_per_oligo()
    expected_barcodes = np.array([[2, 1, 2], [2, 1, 2], [2, 1, 2]])
    np.testing.assert_array_equal(supporting_barcodes.to_numpy(), expected_barcodes)


def test_compute_supporting_barcodes_with_filter(mpra_data_with_bc_filter):
    supporting_barcodes = mpra_data_with_bc_filter._supporting_barcodes_per_oligo()
    expected_barcodes = np.array([[2, 0, 2], [1, 1, 2], [2, 1, 0]])
    np.testing.assert_array_equal(supporting_barcodes.to_numpy(), expected_barcodes)


def test_raw_dna_counts(mpra_data):
    expected_dna_counts = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
    np.testing.assert_array_equal(mpra_data.raw_dna_counts, expected_dna_counts)


def test_raw_dna_counts_with_modification(mpra_data):
    mpra_data.data.layers["dna"] = np.array([[10, 20, 30, 10, 20], [40, 50, 60, 40, 50], [70, 80, 90, 100, 1000]])
    expected_dna_counts = np.array([[10, 20, 30, 10, 20], [40, 50, 60, 40, 50], [70, 80, 90, 100, 1000]])
    np.testing.assert_array_equal(mpra_data.raw_dna_counts, expected_dna_counts)


def test_filtered_dna_counts(mpra_data, mpra_data_with_bc_filter):
    expected_dna_counts = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
    np.testing.assert_array_equal(mpra_data.dna_counts, expected_dna_counts)
    expected_filtered_dna_counts = np.array([[1, 2, 0, 1, 2], [0, 5, 6, 4, 5], [7, 8, 9, 0, 0]])
    np.testing.assert_array_equal(mpra_data_with_bc_filter.dna_counts, expected_filtered_dna_counts)


def test_dna_counts_with_sampling(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.DNA, proportion=0.5)
    dna_sampling = mpra_data.data.layers["dna_sampling"]
    np.testing.assert_array_equal(mpra_data.dna_counts, dna_sampling)


def test_dna_counts_with_filter(mpra_data, mpra_data_with_bc_filter):
    mpra_data.apply_count_sampling(CountSampling.DNA, max_value=2)
    expected_filtered_dna_counts = np.array([[1, 2, 2, 1, 2], [2, 2, 2, 2, 2], [2, 2, 2, 2, 2]])
    np.testing.assert_array_equal(mpra_data.dna_counts, expected_filtered_dna_counts)
    mpra_data_with_bc_filter.apply_count_sampling(CountSampling.DNA, max_value=2)
    expected_filtered_dna_counts = np.array([[1, 2, 0, 1, 2], [0, 2, 2, 2, 2], [2, 2, 2, 0, 0]])
    np.testing.assert_array_equal(mpra_data_with_bc_filter.dna_counts, expected_filtered_dna_counts)


@pytest.fixture
def mpra_data_barcode():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


def test_apply_barcode_filter_min_count(mpra_data_barcode):
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MIN_COUNT, params={"rna_min_count": 4, "dna_min_count": 3})
    expected_filter = np.array(
        [
            [True, False, False],
            [True, False, False],
            [False, False, False],
            [True, False, False],
            [True, False, False],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)


def test_apply_barcode_filter_max_count(mpra_data_barcode):
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MAX_COUNT, params={"rna_max_count": 9, "dna_max_count": 100})
    expected_filter = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, True],
            [False, False, True],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)

    mpra_data_barcode.var_filter = None
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MAX_COUNT, params={"dna_max_count": 99})
    expected_filter = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, True],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)  # type: ignore


@pytest.fixture
def mpra_data_norm():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    data.scaling = 10
    return data


@pytest.fixture
def mpra_data_norm_with_bc_filter(mpra_data_norm):
    data = copy.deepcopy(mpra_data_norm)
    data.var_filter = FILTER
    return data


def test_normalize_counts(mpra_data_norm):
    mpra_data_norm._normalize()
    dna_normalized = mpra_data_norm.normalized_dna_counts
    expected_dna_normalized = np.array(
        [
            [1.428, 2.142, 2.857, 1.428, 2.142],
            [1.724, 2.068, 2.413, 1.724, 2.068],
            [0.575, 0.647, 0.719, 0.791, 7.266],
        ]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array(
        [[1.333, 2.0, 3.333, 1.333, 2.0], [1.724, 2.069, 2.414, 1.724, 2.069], [0.576, 0.647, 0.719, 0.791, 7.266]]
    )
    rna_normalized = mpra_data_norm.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_normalize_without_pseudocount(mpra_data_norm):
    mpra_data = copy.deepcopy(mpra_data_norm)
    mpra_data.pseudo_count = 0
    mpra_data._normalize()
    dna_normalized = np.asarray(mpra_data.data.layers["dna_normalized"])
    expected_dna_normalized = np.array(
        [[1.111, 2.222, 3.333, 1.111, 2.222], [1.667, 2.083, 2.5, 1.667, 2.083], [0.522, 0.597, 0.672, 0.746, 7.463]]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array(
        [[1.0, 2.0, 4.0, 1.0, 2.0], [1.667, 2.083, 2.5, 1.667, 2.083], [0.522, 0.597, 0.672, 0.746, 7.463]]
    )
    rna_normalized = np.asarray(mpra_data.data.layers["rna_normalized"])
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_normalize_counts_with_bc_filter(mpra_data_norm_with_bc_filter):
    mpra_data_norm_with_bc_filter._normalize()
    dna_normalized = np.asarray(mpra_data_norm_with_bc_filter.data.layers["dna_normalized"])
    expected_normalized = np.array([[2.0, 3.0, 0.0, 2.0, 3.0], [0.0, 2.5, 2.917, 2.083, 2.5], [2.963, 3.333, 3.704, 0.0, 0.0]])
    np.testing.assert_almost_equal(dna_normalized, expected_normalized, decimal=3)
    rna_normalized = np.asarray(mpra_data_norm_with_bc_filter.data.layers["rna_normalized"])
    np.testing.assert_almost_equal(rna_normalized, expected_normalized, decimal=3)


@pytest.fixture
def mpra_oligo_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    mpra_barcode_data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    data = mpra_barcode_data.oligo_data
    data.scaling = 10
    return data


@pytest.fixture
def mpra_oligo_data_with_bc_filter():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    mpra_barcode_data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    mpra_barcode_data.var_filter = FILTER
    data = mpra_barcode_data.oligo_data
    data.scaling = 10
    return data


def test_oligo_normalize_counts(mpra_oligo_data):
    dna_normalized = mpra_oligo_data.normalized_dna_counts
    expected_dna_normalized = np.array([[1.786, 2.857, 1.786], [1.897, 2.414, 1.897], [0.612, 0.719, 4.029]])
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array([[1.667, 3.333, 1.667], [1.897, 2.414, 1.897], [0.612, 0.719, 4.029]])
    rna_normalized = mpra_oligo_data.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_oligo_normalize_without_pseudocount(mpra_oligo_data):
    mpra_data = copy.deepcopy(mpra_oligo_data)
    mpra_data.pseudo_count = 0
    dna_normalized = mpra_data.normalized_dna_counts
    expected_dna_normalized = np.array([[1.667, 3.333, 1.667], [1.875, 2.5, 1.875], [0.56, 0.672, 4.104]])
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array([[1.5, 4.0, 1.5], [1.875, 2.5, 1.875], [0.56, 0.672, 4.104]])
    rna_normalized = mpra_data.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_oligo_normalize_counts_with_bc_filter(mpra_oligo_data_with_bc_filter):
    dna_normalized = mpra_oligo_data_with_bc_filter.normalized_dna_counts
    expected_normalized = np.array([[2.5, 0.0, 2.5], [2.5, 2.917, 2.292], [3.148, 3.704, 0.0]])
    np.testing.assert_almost_equal(dna_normalized, expected_normalized, decimal=3)
    rna_normalized = mpra_oligo_data_with_bc_filter.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_normalized, decimal=3)


@pytest.fixture
def mpra_corr_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers)).oligo_data


def test_correlation(mpra_corr_data):
    mpra_corr_data._compute_correlation(mpra_corr_data.activity, "log2FoldChange")
    assert "pearson_correlation_log2FoldChange" in mpra_corr_data.data.obsp
    assert "spearman_correlation_log2FoldChange" in mpra_corr_data.data.obsp


def test_pearson_correlation(mpra_corr_data):
    x = mpra_corr_data.correlation(method="pearson", count_type=Modality.ACTIVITY)
    y = mpra_corr_data.correlation(method="pearson", count_type=Modality.RNA_NORMALIZED)
    z = mpra_corr_data.correlation(method="pearson", count_type=Modality.DNA_NORMALIZED)
    np.testing.assert_equal(x, np.array([[1.0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]))
    np.testing.assert_almost_equal(
        y, np.array([[1.0, 1.0, -0.475752], [1.0, 1.0, -0.475752], [-0.475752, -0.475752, 1.0]]), decimal=3
    )
    np.testing.assert_almost_equal(z, np.array([[1.0, 1.0, -0.476], [1.0, 1.0, -0.476], [-0.476, -0.476, 1.0]]), decimal=3)


def test_spearman_correlation(mpra_corr_data):
    x = mpra_corr_data.correlation(method="spearman", count_type=Modality.ACTIVITY)
    y = mpra_corr_data.correlation(method="spearman", count_type=Modality.RNA_NORMALIZED)
    z = mpra_corr_data.correlation(method="spearman", count_type=Modality.DNA_NORMALIZED)
    np.testing.assert_equal(x, np.array([[1.0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]))
    np.testing.assert_almost_equal(y, np.array([[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]]), decimal=3)
    np.testing.assert_equal(z, np.array([[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]]))


@pytest.fixture
def mpra_complexity_data():
    counts_dna = np.array([[0, 2, 0, 1, 2], [4, 5, 6, 4, 5], [7, 0, 9, 10, 0]])
    counts_rna = np.array([[1, 2, 0, 1, 2], [4, 5, 6, 4, 5], [7, 0, 9, 10, 0]])
    layers = {"rna": counts_rna, "dna": counts_dna}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


def test_lincoln_complexity(mpra_complexity_data):
    complexity = mpra_complexity_data.complexity()
    np.testing.assert_equal(complexity, np.array([[4, 5, 6], [5, 5, 5], [6, 5, 3]]))


def test_chapman_complexity(mpra_complexity_data):
    complexity = mpra_complexity_data.complexity(method="chapman")
    np.testing.assert_equal(complexity, np.array([[4, 5, 5], [5, 5, 5], [5, 5, 3]]))


def test_fail_complexity(mpra_complexity_data):
    with pytest.raises(ValueError):
        mpra_complexity_data.complexity(method="unknown")


def test_read_and_write(tmp_path, mpra_data):
    out_path = tmp_path / "bc_data.h5ad"
    mpra_data.write(out_path)

    data = MPRABarcodeData.read(out_path)
    assert isinstance(data, MPRABarcodeData)
    assert data.data.shape == mpra_data.data.shape
    assert data.data.layers.keys() == mpra_data.data.layers.keys()
    assert np.all(data.rna_counts == mpra_data.rna_counts)
    assert np.all(data.activity == mpra_data.activity)
    assert data.pseudo_count == mpra_data.pseudo_count
    assert data.scaling == mpra_data.scaling


def test_read_and_write_oligo(tmp_path, mpra_oligo_data):
    out_path = tmp_path / "oligo_data.h5ad"
    mpra_oligo_data.write(out_path)

    data = MPRAOligoData.read(out_path)
    assert isinstance(data, MPRAOligoData)
    assert data.data.shape == mpra_oligo_data.data.shape
    assert data.data.layers.keys() == mpra_oligo_data.data.layers.keys()
    assert np.all(data.rna_counts == mpra_oligo_data.rna_counts)
    assert np.all(data.activity == mpra_oligo_data.activity)


def test_read_and_write_with_modifications(tmp_path, mpra_data):
    out_path = tmp_path / "bc_data_mod.h5ad"
    mpra_data.scaling = 10.0
    mpra_data.pseudo_count = 0
    mpra_data.write(out_path)

    data = MPRABarcodeData.read(out_path)
    assert isinstance(data, MPRABarcodeData)
    assert data.scaling == 10.0
    assert data.pseudo_count == 0



import os
import tempfile
import pandas as pd
import numpy as np
import pytest
import anndata as ad
from mpralib.utils.io import read_sequence_design_file, export_counts_file, chromosome_map
from mpralib.exception import SequenceDesignException, MPRAlibException
from mpralib.mpradata import MPRABarcodeData, MPRAOligoData, MPRAData


class DummyMPRAData(MPRAData):
    def __init__(self, replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold, barcode_counts=None):
        layers = {"rna": rna_counts, "dna": dna_counts}
        obs = pd.DataFrame(index=replicates)
        var = pd.DataFrame({"oligo": oligos}, index=barcodes)
        super().__init__(ad.AnnData(X=rna_counts, obs=obs, var=var, layers=layers), barcode_threshold)
        if barcode_counts:
            self.barcode_counts = pd.DataFrame(
                barcode_counts,
                index=self.obs_names,
                columns=self.var_names,
            )


class DummyMPRABarcodeData(DummyMPRAData, MPRABarcodeData):
    pass


class DummyMPRAOligoData(DummyMPRAData, MPRAOligoData):
    pass


@pytest.fixture
def barcode_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]])
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]])
    barcode_threshold = 1
    return DummyMPRABarcodeData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold)


@pytest.fixture
def oligo_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]])
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]])
    barcode_counts = [[1, 0, 1], [1, 0, 1]]
    barcode_threshold = 1
    return DummyMPRAOligoData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold, barcode_counts)


def test_export_counts_file_barcode(tmp_path, barcode_data):
    out_path = tmp_path / "counts.tsv"
    export_counts_file(barcode_data, str(out_path), normalized=False)
    df = pd.read_csv(out_path, sep="\t", na_values=["X"])
    # Only rows with observed=True and barcode_counts >= threshold should remain
    assert "oligo_name" in df.columns
    assert set(df["oligo_name"]) == {"ol1", "ol2", "ol3"}
    assert "dna_count_rep1" in df.columns
    assert "rna_count_rep2" in df.columns
    assert "rna_count_rep2" in df.columns
    assert df.loc[df["oligo_name"] == "ol3", "dna_count_rep1"].values[0] == 5
    # Check that zeros are written as empty
    assert np.isnan(df.loc[df["oligo_name"] == "ol3", "dna_count_rep2"].values[0])


def test_export_counts_file_oligo(tmp_path, oligo_data):
    out_path = tmp_path / "counts.tsv"
    export_counts_file(oligo_data, str(out_path), normalized=False)
    df = pd.read_csv(out_path, sep="\t")
    # Only rows with observed=True and barcode_counts >= threshold should remain
    assert "ID" in df.columns
    assert set(df["ID"]) == {"ol1", "ol3"}
    assert "dna_count_rep1" in df.columns
    assert "rna_count_rep2" in df.columns


def test_export_counts_file_normalized(tmp_path, barcode_data):
    out_path = tmp_path / "counts_norm.tsv"
    export_counts_file(barcode_data, str(out_path), normalized=True)
    df = pd.read_csv(out_path, sep="\t")
    # Check float formatting
    assert np.all(df.filter(like="dna_count").applymap(lambda x: isinstance(x, float) or np.isnan(x)).values)  # type: ignore


def test_export_counts_file_with_filter(tmp_path, barcode_data):
    out_path = tmp_path / "counts_filt.tsv"
    # Mask out the first entry
    filter_mask = np.asarray([[True, False, False], [True, False, False]])
    export_counts_file(barcode_data, str(out_path), normalized=False, filter=filter_mask)
    df = pd.read_csv(out_path, sep="\t")
    # Only ol3 should remain (ol2 fails barcode_threshold, ol1 is masked)
    assert set(df["oligo_name"]) == {"ol2", "ol3"}


def test_export_counts_file_invalid_type(tmp_path):
    class Dummy:
        pass

    dummy = Dummy()
    with pytest.raises(MPRAlibException):
        export_counts_file(dummy, str(tmp_path / "fail.tsv"))  # type: ignore


VALID_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["ref"]\tinfo1
oligo2\tGGCATGCA\telement\tvariant positive control\tsrc2\tref2\tchr2\t300\t400\t-\tNA\tNA\tNA\tNA\tinfo2
oligo3\tATGCATGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""

INVALID_SEQ_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCXTGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""

INVALID_CATEGORY_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tbadcat\ttest\tsrc1\tref1\tchr1\t100\t200\t+\tNA\tNA\tNA\tNA\tinfo1
"""

INVALID_CLASS_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tvariant\tbadclass\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""


def write_temp_tsv(content):
    fd, path = tempfile.mkstemp(suffix=".tsv")
    with os.fdopen(fd, "w") as f:
        f.write(content)
    return path


def test_read_sequence_design_file_valid():
    path = write_temp_tsv(VALID_TSV)
    df = read_sequence_design_file(path)
    assert isinstance(df, pd.DataFrame)
    assert "sequence" in df.columns
    assert df.shape[0] == 3
    assert set(df.index) == {"oligo1", "oligo2", "oligo3"}
    os.remove(path)


def test_read_sequence_design_file_invalid_sequence():
    path = write_temp_tsv(INVALID_SEQ_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "sequence" in str(excinfo.value)
    os.remove(path)


def test_read_sequence_design_file_invalid_category():
    path = write_temp_tsv(INVALID_CATEGORY_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "category" in str(excinfo.value)
    os.remove(path)


def test_read_sequence_design_file_invalid_class():
    path = write_temp_tsv(INVALID_CLASS_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "class" in str(excinfo.value)
    os.remove(path)


def test_chromosome_map_columns():
    df = chromosome_map()
    expected_columns = ["ucsc", "assembly", "genbank", "refseq", "release"]
    assert list(df.columns) == expected_columns


def test_chromosome_map_release_values():
    df = chromosome_map()
    assert set(df["release"].unique()) == {"GRCh37", "GRCh38"}


def test_chromosome_map_non_empty():
    df = chromosome_map()
    assert not df.empty
    assert df.shape[0] > 0


def test_chromosome_map_no_duplicate_rows():
    df = chromosome_map()
    assert df.duplicated().sum() == 0



import unittest
import os
import tempfile
from click.testing import CliRunner
from mpralib.cli import cli, _get_chr
import gzip
import pandas as pd
from logging import Logger


class TestMPRlibCLI(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

        # Create a temporary input file
        self.input_file = os.path.join(
            os.path.dirname(__file__),
            "data",
            "reporter_experiment_barcode.input.tsv.gz",
        )

    def test_barcode_activities_bc1(self):

        # Create a temporary output file
        with tempfile.NamedTemporaryFile(delete=False) as temp_output:
            output_file = temp_output.name

        # Run the command
        result = self.runner.invoke(
            cli,
            [
                "functional",
                "activities",
                "--input",
                self.input_file,
                "--barcode-level",
                "--output",
                output_file,
            ],
        )

        # Check the result
        self.assertIs(result.exit_code, 0)
        self.assertTrue(os.path.exists(output_file))

        with open(output_file, "r") as f:
            output_content = f.read()

        expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")

        with gzip.open(expected_output_file, "rt") as f:
            expected_content = f.read()

        self.assertTrue(output_content == expected_content)

    def test_activities_bc1(self):

        # Create a temporary output file
        with tempfile.NamedTemporaryFile(delete=False) as temp_output:
            output_file = temp_output.name

        # Run the command
        result = self.runner.invoke(
            cli,
            [
                "functional",
                "activities",
                "--input",
                self.input_file,
                "--bc-threshold",
                "1",
                "--output",
                output_file,
            ],
        )

        # Check the result
        self.assertIs(result.exit_code, 0)
        self.assertTrue(os.path.exists(output_file))

        with open(output_file, "r") as f:
            output_content = f.read()

        expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc1.output.tsv.gz")

        with gzip.open(expected_output_file, "rt") as f:
            expected_content = f.read()

        self.assertTrue(output_content == expected_content)

    def test_activities_bc10(self):

        # Create a temporary output file
        with tempfile.NamedTemporaryFile(delete=False) as temp_output:
            output_file = temp_output.name

        # Run the command
        result = self.runner.invoke(
            cli,
            [
                "functional",
                "activities",
                "--input",
                self.input_file,
                "--bc-threshold",
                "10",
                "--output",
                output_file,
            ],
        )

        # Check the result
        assert result.exit_code == 0
        assert os.path.exists(output_file)

        with open(output_file, "r") as f:
            output_content = f.read()

        expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc10.output.tsv.gz")

        with gzip.open(expected_output_file, "rt") as f:
            expected_content = f.read()

        assert output_content == expected_content

    def test_activities_bc100(self):

        # Create a temporary output file
        with tempfile.NamedTemporaryFile(delete=False) as temp_output:
            output_file = temp_output.name

        # Run the command
        result = self.runner.invoke(
            cli,
            [
                "functional",
                "activities",
                "--input",
                self.input_file,
                "--bc-threshold",
                "100",
                "--output",
                output_file,
            ],
        )

        # Check the result
        assert result.exit_code == 0
        assert os.path.exists(output_file)

        with open(output_file, "r") as f:
            output_content = f.read()

        expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc100.output.tsv.gz")

        with gzip.open(expected_output_file, "rt") as f:
            expected_content = f.read()

        assert output_content == expected_content

        # Clean up
        os.remove(output_file)


class DummyLogger(Logger):

    def __init__(self):
        self.messages = []

    def warning(self, msg, *args, **kwargs):
        self.messages.append(msg)


def test_get_chr_found():
    # Prepare a chromosome map DataFrame
    map_df = pd.DataFrame({
        "refseq": ["NC_000001.11", "NC_000002.12"],
        "ucsc": ["chr1", "chr2"],
        "release": ["GRCh38", "GRCh38"]
    })
    variant_id = "NC_000001.11:12345:A:T"
    logger = DummyLogger()
    result = _get_chr(map_df, variant_id, logger)
    assert result == "chr1"
    assert logger.messages == []


def test_get_chr_not_found():
    map_df = pd.DataFrame({
        "refseq": ["NC_000001.11", "NC_000002.12"],
        "ucsc": ["chr1", "chr2"],
        "release": ["GRCh38", "GRCh38"]
    })
    variant_id = "NC_000003.13:54321:G:C"
    logger = DummyLogger()
    result = _get_chr(map_df, variant_id, logger)
    assert result is None
    assert any("Contig NC_000003.13 of SPDI NC_000003.13:54321:G:C not found" in msg for msg in logger.messages)


def test_get_chr_handles_empty_map():
    map_df = pd.DataFrame(columns=["refseq", "ucsc", "release"])
    variant_id = "NC_000004.14:11111:T:A"
    logger = DummyLogger()
    result = _get_chr(map_df, variant_id, logger)
    assert result is None
    assert any("Contig NC_000004.14 of SPDI NC_000004.14:11111:T:A not found" in msg for msg in logger.messages)


def test_get_chr_with_multiple_matches():
    # Should return the first match if multiple rows match
    map_df = pd.DataFrame({
        "refseq": ["NC_000005.15", "NC_000005.15"],
        "ucsc": ["chr5a", "chr5b"],
        "release": ["GRCh38", "GRCh37"]
    })
    variant_id = "NC_000005.15:22222:C:G"
    logger = DummyLogger()
    result = _get_chr(map_df, variant_id, logger)
    assert result in ["chr5a", "chr5b"]
    assert logger.messages == []



import numpy as np
import pandas as pd
import pytest
import anndata as ad
import seaborn as sns
import copy
from mpralib.utils.plot import dna_vs_rna, correlation, barcodes_per_oligo
from mpralib.mpradata import MPRAData, MPRABarcodeData, MPRAOligoData, Modality


class DummyMPRAData(MPRAData):
    def __init__(self, replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold, barcode_counts=None):
        layers = {"rna": rna_counts, "dna": dna_counts}
        obs = pd.DataFrame(index=replicates)
        var = pd.DataFrame({"oligo": oligos}, index=barcodes)
        super().__init__(ad.AnnData(X=rna_counts, obs=obs, var=var, layers=layers), barcode_threshold)
        if barcode_counts:
            self.barcode_counts = pd.DataFrame(
                barcode_counts,
                index=self.obs_names,
                columns=self.var_names,
            )


class DummyMPRABarcodeData(DummyMPRAData, MPRABarcodeData):
    pass


class DummyMPRAOligoData(DummyMPRAData, MPRAOligoData):
    pass


@pytest.fixture
def barcode_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]])
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]])
    barcode_threshold = 1
    return copy.deepcopy(DummyMPRABarcodeData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold))


@pytest.fixture
def oligo_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]])
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]])
    barcode_counts = [[1, 0, 1], [1, 0, 1]]
    barcode_threshold = 1
    return copy.deepcopy(
        DummyMPRAOligoData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold, barcode_counts)
    )


@pytest.mark.parametrize(
    "layer",
    [
        Modality.DNA,
        Modality.RNA,
        Modality.RNA_NORMALIZED,
        Modality.DNA_NORMALIZED,
        Modality.ACTIVITY,
    ],
)
def test_correlation_returns_pairgrid(oligo_data, layer):
    g = correlation(oligo_data, layer)
    assert isinstance(g, sns.PairGrid)
    # Should have as many axes as replicates
    n = len(oligo_data.obs_names)
    assert g.axes.shape == (n, n)


def test_correlation_with_replicates_subset(oligo_data):
    g = correlation(oligo_data, Modality.DNA, replicates=["rep1", "rep2"])
    assert isinstance(g, sns.PairGrid)
    assert g.axes.shape == (2, 2)
    # Check that the columns are correct
    colnames = [ax.get_xlabel() for ax in g.axes[0]]
    assert any("Replicate" in c for c in colnames)


def test_correlation_masking(oligo_data):
    # Set barcode_counts below threshold for one oligo
    oligo_data.barcode_counts[0, 2] = 0
    g = correlation(oligo_data, Modality.DNA)
    # Should still return a PairGrid
    assert isinstance(g, sns.PairGrid)


@pytest.mark.parametrize(
    "layer",
    [
        Modality.DNA,
        Modality.RNA,
        Modality.RNA_NORMALIZED,
        Modality.DNA_NORMALIZED,
        Modality.ACTIVITY,
    ],
)
def test_correlation_returns_pairgrid_barcode(barcode_data, layer):
    g = correlation(barcode_data, layer)
    assert isinstance(g, sns.PairGrid)
    # Should have as many axes as replicates
    n = len(barcode_data.obs_names)
    assert g.axes.shape == (n, n)


def test_correlation_with_replicates_subset_barcode(barcode_data):
    g = correlation(barcode_data, Modality.DNA, replicates=["rep1", "rep2"])
    assert isinstance(g, sns.PairGrid)
    assert g.axes.shape == (2, 2)
    # Check that the columns are correct
    colnames = [ax.get_xlabel() for ax in g.axes[0]]
    assert any("Replicate" in c for c in colnames)


def test_dna_vs_rna_oligo_returns_jointgrid(oligo_data):
    g = dna_vs_rna(oligo_data)
    assert isinstance(g, sns.axisgrid.JointGrid)
    # Check axis labels
    assert g.ax_joint.get_xlabel() == "DNA [log10]"
    assert g.ax_joint.get_ylabel() == "RNA [log10]"


def test_dna_vs_rna_barcode_returns_jointgrid(barcode_data):
    g = dna_vs_rna(barcode_data)
    assert isinstance(g, sns.axisgrid.JointGrid)
    # Check axis labels
    assert g.ax_joint.get_xlabel() == "DNA [log10]"
    assert g.ax_joint.get_ylabel() == "RNA [log10]"


def test_dna_vs_rna_oligo_with_replicates(oligo_data):
    g = dna_vs_rna(oligo_data, replicates=["rep2"])
    assert isinstance(g, sns.axisgrid.JointGrid)
    # Check axis labels
    assert g.ax_joint.get_xlabel() == "DNA [log10]"
    assert g.ax_joint.get_ylabel() == "RNA [log10]"


def test_barcodes_per_oligo(oligo_data):
    g = barcodes_per_oligo(oligo_data)
    assert isinstance(g, sns.axisgrid.FacetGrid)
    assert g.axes.shape == (1, 2)


def test_barcodes_per_oligo_with_replicates(oligo_data):
    g = barcodes_per_oligo(oligo_data, replicates=["rep2"])
    assert isinstance(g, sns.axisgrid.FacetGrid)
    assert g.axes.shape == (1, 1)
